# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
These routines perform some hierrachical agglomerative clustering
of some input data. The following alternatives are proposed:
- Distance based average-link
- Similarity-based average-link
- Distance based maximum-link
- Ward's algorithm under graph constraints
- Ward's algorithm without graph constraints

In this latest version, the results are returned in a 'WeightedForest'
structure, which gives access to the clustering hierarchy, facilitates
the plot of the result etc.

For back-compatibility, *_segment versions of the algorithms have been
appended, with the old API (except the qmax parameter, which now
represents the number of wanted clusters)

Author : Bertrand Thirion,Pamela Guevara, 2006-2009
"""

#---------------------------------------------------------------------------
# ------ Routines for Agglomerative Hierarchical Clustering ----------------
# --------------------------------------------------------------------------

import numpy as np

from ..graph.graph import WeightedGraph
from ..graph.forest import Forest


class WeightedForest(Forest):
    """
    This is a weighted Forest structure, i.e. a tree
    - each node has one parent and children
    (hierarchical structure)
    - some of the nodes can be viewed as leaves, other as roots
    - the edges within a tree are associated with a weight:
    +1 from child to parent
    -1 from parent to child
    - additionally, the nodes have a value, which is called 'height',
    especially useful from dendrograms

    members
    -------
    V : (int, >0) the number of vertices
    E : (int) the number of edges
    parents: array of shape (self.V) the parent array
    edges: array of shape (self.E,2) reprensenting pairwise neighbors
    weights, array of shape (self.E), +1/-1 for scending/descending links
    children: list of arrays that represents the childs of any node
    height: array of shape(self.V)
    """

    def __init__(self, V, parents=None, height=None):
        """
        Parameters
        ----------
        V: the number of edges of the graph
        parents=None: array of shape (V)
                the parents of the graph
                by default, the parents are set to range(V), i.e. each
                node is its own parent, and each node is a tree
        height=None: array of shape(V)
                     the height of the nodes
        """
        V = int(V)
        if V < 1:
            raise ValueError('cannot create graphs with no vertex')
        self.V = int(V)

        # define the parents
        if parents == None:
            self.parents = np.arange(self.V)
        else:
            if np.size(parents) != V:
                raise ValueError('Incorrect size for parents')
            if parents.max() > self.V:
                raise ValueError('Incorrect value for parents')
            self.parents = np.reshape(parents, self.V)

        self.define_graph_attributes()

        if self.check() == 0:
            raise ValueError('The proposed structure is not a forest')
        self.children = []

        if height == None:
            height = np.zeros(self.V)
        else:
            if np.size(height) != V:
                raise ValueError('Incorrect size for height')
            self.height = np.reshape(height, self.V)

    def set_height(self, height=None):
        """Set the height array
        """
        if height == None:
            height = np.zeros(self.V)

        if np.size(height) != self.V:
            raise ValueError('Incorrect size for height')

        self.height = np.reshape(height, self.V)

    def get_height(self):
        """Get the height array
        """
        return self.height

    def check_compatible_height(self):
        """Check that height[parents[i]]>=height[i] for all nodes
        """
        OK = True
        for i in range(self.V):
            if self.height[self.parents[i]] < self.height[i]:
                OK = False
        return OK

    def plot(self, ax=None):
        """Plot the dendrogram associated with self
        the rank of the data in the dendogram is returned

        Parameters
        ----------
        ax: axis handle, optional

        Returns
        -------
        ax, the axis handle
        """
        import matplotlib.pylab as mp
        if self.check_compatible_height() == False:
            raise ValueError('cannot plot myself in my current state')

        n = np.sum(self.isleaf())

        # 1. find a permutation of the leaves that makes it nice
        aux = _label(self.parents)
        temp = np.zeros(self.V)
        rank = np.arange(self.V)
        temp[:n] = np.argsort(aux[:n])
        for i in range(n):
            rank[temp[i]] = i

        # 2. derive the abscissa in the dendrogram
        idx = np.zeros(self.V)
        temp = np.argsort(rank[:n])
        for i in range(n):
            idx[temp[i]] = i
        for i in range(n, self.V):
            j = np.nonzero(self.parents == i)[0]
            idx[i] = np.mean(idx[j])

        # 3. plot
        if ax == None:
            mp.figure()
            ax = mp.subplot(1, 1, 1)

        for i in range(self.V):
            h1 = self.height[i]
            h2 = self.height[self.parents[i]]
            mp.plot([idx[i], idx[i]], [h1, h2], 'k')

        ch = self.get_children()
        for i in range(self.V):
            if np.size(ch[i]) > 0:
                lidx = idx[ch[i]]
                m = lidx.min()
                M = lidx.max()
                h = self.height[i]
                mp.plot([m, M], [h, h], 'k')

        cM = 1.05 * self.height.max() - 0.05 * self.height.min()
        cm = 1.05 * self.height.min() - 0.05 * self.height.max()
        mp.axis([-1, idx.max() + 1, cm, cM])
        return ax

    def partition(self, threshold):
        """ Partition the tree according to a cut criterion
        """
        valid = self.height < threshold
        f = self.subforest(valid)
        u = f.cc()
        return u[f.isleaf()]

    def split(self, k):
        """
        idem as partition, but a number of components are supplied instead
        """
        k = int(k)
        if k > self.V:
            k = self.V
        nbcc = self.cc().max() + 1

        if k <= nbcc:
            u = self.cc()
            return u[self.isleaf()]

        sh = np.sort(self.height)
        th = sh[nbcc - k]
        u = self.partition(th)
        return u

    def plot_height(self):
        """Plot the height of the non-leaves nodes
        """
        import matplotlib.pylab as mp
        mp.figure()
        sh = np.sort(self.height[self.isleaf() == False])
        n = np.sum(self.isleaf() == False)
        mp.bar(np.arange(n), sh)

    def list_of_subtrees(self):
        """
        returns the list of all non-trivial subtrees in the graph
        Caveat: theis function assumes that the vertices are sorted in a
        way such that parent[i]>i forall i
        Only the leaves are listeed, not the subtrees themselves
        """
        lst = []
        n = np.sum(self.isleaf())
        for i in range(self.V):
            lst.append(np.array([], np.int))
        for i in range(n):
            lst[i] = np.array([i], np.int)
        for i in range(self.V - 1):
            j = self.parents[i]
            lst[j] = np.hstack((lst[i], lst[j]))

        return lst[n:self.V]


#--------------------------------------------------------------------------
#------------- Average link clustering on a graph -------------------------
# -------------------------------------------------------------------------


def fusion(K, pop, i, j, k):
    """ Modifies the graph K to merge nodes i and  j into nodes k

    The similarity values are weighted averaged, where pop[i] and pop[j]
    yield the relative weights.
    this is used in average_link_slow (deprecated)
    """
    #
    fi = float(pop[i]) / (pop[k])
    fj = 1.0 - fi
    #
    # replace i ny k
    #
    idxi = np.nonzero(K.edges[:, 0] == i)
    K.weights[idxi] = K.weights[idxi] * fi
    K.edges[idxi, 0] = k
    idxi = np.nonzero(K.edges[:, 1] == i)
    K.weights[idxi] = K.weights[idxi] * fi
    K.edges[idxi, 1] = k
    #
    # replace j by k
    #
    idxj = np.nonzero(K.edges[:, 0] == j)
    K.weights[idxj] = K.weights[idxj] * fj
    K.edges[idxj, 0] = k
    idxj = np.nonzero(K.edges[:, 1] == j)
    K.weights[idxj] = K.weights[idxj] * fj
    K.edges[idxj, 1] = k
    #
    #sum/remove double edges
    #
    #left side
    idxk = np.nonzero(K.edges[:, 0] == k)[0]
    corr = K.edges[idxk, 1]
    scorr = np.sort(corr)
    acorr = np.argsort(corr)
    for a in range(np.size(scorr) - 1):
        if scorr[a] == scorr[a + 1]:
            i1 = idxk[acorr[a]]
            i2 = idxk[acorr[a + 1]]
            K.weights[i1] = K.weights[i1] + K.weights[i2]
            K.weights[i2] = - np.inf
            K.edges[i2] = -1

    #right side
    idxk = np.nonzero(K.edges[:, 1] == k)[0]
    corr = K.edges[idxk, 0]
    scorr = np.sort(corr)
    acorr = np.argsort(corr)
    for a in range(np.size(scorr) - 1):
        if scorr[a] == scorr[a + 1]:
            i1 = idxk[acorr[a]]
            i2 = idxk[acorr[a + 1]]
            K.weights[i1] = K.weights[i1] + K.weights[i2]
            K.weights[i2] = - np.inf
            K.edges[i2] = - 1


def average_link_graph(G):
    """
    Agglomerative function based on a (hopefully sparse) similarity graph

    Parameters
    ----------
    G the input graph

    Returns
    -------
    t a weightForest structure that represents the dendrogram of the data

    CAVEAT
    ------
    In that case, the homogeneity is associated with high similarity
    (as opposed to low cost as in most clustering procedures,
    e.g. distance-based procedures).  Thus the tree is created with
    negated affinity values, in roder to respect the traditional
    ordering of cluster potentials. individual points have the
    potential (-np.inf).
    This problem is handled transparently inthe associated segment functionp.
    """

    # prepare a graph with twice the number of vertices
    n = G.V
    nbcc = G.cc().max() + 1
    K = WeightedGraph(2 * G.V)
    K.E = G.E
    K.edges = G.edges.copy()
    K.weights = G.weights.copy()

    parent = np.arange(2 * n - nbcc, dtype=np.int)
    pop = np.ones(2 * n - nbcc, np.int)
    height = np.inf * np.ones(2 * n - nbcc)

    # iteratively merge clusters
    for q in range(n - nbcc):

        # 1. find the heaviest edge
        m = (K.weights).argmax()
        cost = K.weights[m]
        k = q + n
        height[k] = cost
        i = K.edges[m, 0]
        j = K.edges[m, 1]

        # 2. remove the current edge
        K.edges[m] = -1
        K.weights[m] = - np.inf
        m = np.nonzero((K.edges[:, 0] == j) * (K.edges[:, 1] == i))[0]
        K.edges[m] = - 1
        K.weights[m] = - np.inf

        # 3. merge the edges with third part edges
        parent[i] = k
        parent[j] = k
        pop[k] = pop[i] + pop[j]
        fusion(K, pop, i, j, k)

    height[height < 0] = 0
    height[np.isinf(height)] = height[n] + 1
    t = WeightedForest(2 * n - nbcc, parent, - height)
    return t


def average_link_graph_segment(G, stop=0, qmax=1, verbose=False):
    """Agglomerative function based on a (hopefully sparse) similarity graph

    Parameters
    ----------
    G the input graph
    stop: float
        the stopping criterion
    qmax: int, optional
        the number of desired clusters (in the limit of the stopping criterion)
    verbose : bool, optional
        If True, print diagnostic information

    Returns
    -------
    u: array of shape (G.V)
       a labelling of the graph vertices according to the criterion
    cost: array of shape (G.V (?))
          the cost of each merge step during the clustering procedure
    """
    # prepare a graph with twice the number of vertices
    n = G.V
    if qmax == - 1:
        qmax = n
    qmax = int(np.minimum(qmax, n))

    t = average_link_graph(G)

    if verbose:
        t.plot()

    u1 = np.zeros(n, np.int)
    u2 = np.zeros(n, np.int)
    if stop >= 0:
        u1 = t.partition( - stop)
    if qmax > 0:
        u2 = t.split(qmax)

    if u1.max() < u2.max():
        u = u2
    else:
        u = u1

    cost = - t.get_height()
    cost = cost[t.isleaf() == False]

    return u, cost


#--------------------------------------------------------------------------
#------------- Ward's algorithm with graph constraints --------------------
# -------------------------------------------------------------------------


def _inertia_(i, j, Features):
    """
    Compute the variance of the set which is
    the concatenation of Feature[i] and Features[j]
    """
    if np.size(np.shape(Features[i])) < 2:
        print i, np.shape(Features[i]), Features[i]
    if np.size(np.shape(Features[i])) < 2:
        print j, np.shape(Features[j]), Features[j]
    if np.shape(Features[i])[1] != np.shape(Features[j])[1]:
        print i, j, np.shape(Features[i]), np.shape(Features[j])
    localset = np.vstack((Features[i], Features[j]))
    return np.var(localset, 0).sum()


def _inertia(i, j, Features):
    """
    Compute the variance of the set which is
    the concatenation of Feature[i] and Features[j]
    """
    n = Features[0][i] + Features[0][j]
    s = Features[1][i] + Features[1][j]
    q = Features[2][i] + Features[2][j]
    return np.sum(q - (s ** 2 / n))


def _initial_inertia(K, Features, seeds=None):
    """ Compute the variance associated with each
    edge-related pair of vertices
    Thre sult is written in K;weights
    if seeds if provided (seeds!=None)
    this is done only for vertices adjacent to the seeds
    """
    if seeds == None:
        for e in range(K.E):
            i = K.edges[e, 0]
            j = K.edges[e, 1]
            ESS = _inertia(i, j, Features)
            K.weights[e] = ESS
    else:
        aux = np.zeros(K.V).astype('bool')
        aux[seeds] = 1
        for e in range(K.E):
            i = K.edges[e, 0]
            j = K.edges[e, 1]
            if (aux[i] or aux[j]):
                K.weights[e] = _inertia(i, j, Features)
            else:
                K.weights[e] = np.inf


def _auxiliary_graph(G, Features):
    """
    prepare a graph with twice the number of vertices
    this graph will contain the connectivity information
    along the merges.
    """
    K = WeightedGraph(2 * G.V - 1)
    K.E = G.E
    K.edges = G.edges.copy()
    K.weights = np.ones(K.E)
    K.symmeterize()
    if K.E > 0:    
        valid = K.edges[:, 0] < K.edges[:, 1]
        K.remove_edges(valid)
    #
    K.remove_trivial_edges()
    _initial_inertia(K, Features)
    return K


def _remap(K, i, j, k, Features, linc, rinc):
    """Modifies the graph K to merge nodes i and  j into nodes k
    the graph weights are modified accordingly

    Parameters
    ----------
    K graph instance:
      the existing graphical model
    i,j,k: int
           indexes of the nodes to be merged and of the parent respectively
    Features: list of node-per-node features
    linc: array of shape(K.V)
          left incidence matrix
    rinc: array of shape(K.V)
          right incidencematrix
    """
    # -------
    # replace i by k
    # --------
    idxi = np.array(linc[i]).astype(np.int)
    if np.size(idxi) > 1:
        for l in idxi:
            K.weights[l] = _inertia(k, K.edges[l, 1], Features)
    elif np.size(idxi) == 1:
        K.weights[idxi] = _inertia(k, K.edges[idxi, 1], Features)
    if np.size(idxi) > 0:
        K.edges[idxi, 0] = k

    idxi = np.array(rinc[i]).astype(np.int)
    if np.size(idxi) > 1:
        for l in idxi:
            K.weights[l] = _inertia(K.edges[l, 0], k, Features)
    elif np.size(idxi) == 1:
        K.weights[idxi] = _inertia(K.edges[idxi, 0], k, Features)
    if np.size(idxi) > 0:
        K.edges[idxi, 1] = k

    #------
    # replace j by k
    #-------
    idxj = np.array(linc[j]).astype(np.int)
    if np.size(idxj) > 1:
        for l in idxj:
            K.weights[l] = _inertia(k, K.edges[l, 1], Features)
    elif np.size(idxj) == 1:
        K.weights[idxj] = _inertia(k, K.edges[idxj, 1], Features)
    if np.size(idxj) > 0:
        K.edges[idxj, 0] = k

    idxj = np.array(rinc[j]).astype(np.int)
    if np.size(idxj) > 1:
        for l in idxj:
            K.weights[l] = _inertia(k, K.edges[l, 0], Features)
    elif np.size(idxj) == 1:
        K.weights[idxj] = _inertia(k, K.edges[idxj, 0], Features)
    if np.size(idxj) > 0:
        K.edges[idxj, 1] = k

    #------
    # update linc,rinc
    #------
    lidxk = list(np.concatenate((linc[j], linc[i])))
    for l in lidxk:
        if K.edges[l, 1] == - 1:
            lidxk.remove(l)

    linc[k] = lidxk
    linc[i] = []
    linc[j] = []
    ridxk = list(np.concatenate((rinc[j], rinc[i])))
    for l in ridxk:
        if K.edges[l, 0] == - 1:
            ridxk.remove(l)

    rinc[k] = ridxk
    rinc[i] = []
    rinc[j] = []

    #------
    #remove double edges
    #------
    #left side
    idxk = np.array(linc[k]).astype(np.int)
    if np.size(idxk) > 0:
        corr = K.edges[idxk, 1]
        scorr = np.sort(corr)
        acorr = np.argsort(corr)
        for a in range(np.size(scorr) - 1):
            if scorr[a] == scorr[a + 1]:
                i2 = idxk[acorr[a + 1]]
                K.weights[i2] = np.inf
                rinc[K.edges[i2, 1]].remove(i2)
                K.edges[i2] = - 1
                linc[k].remove(i2)

    #right side
    idxk = np.array(rinc[k]).astype(np.int)
    if np.size(idxk) > 0:
        corr = K.edges[idxk, 0]
        scorr = np.sort(corr)
        acorr = np.argsort(corr)
        for a in range(np.size(scorr) - 1):
            if scorr[a] == scorr[a + 1]:
                i2 = idxk[acorr[a + 1]]
                K.weights[i2] = np.inf
                linc[K.edges[i2, 0]].remove(i2)
                K.edges[i2] = - 1
                rinc[k].remove(i2)
    return linc, rinc


def ward_quick(G, feature, verbose=False):
    """ Agglomerative function based on a topology-defining graph
    and a feature matrix.

    Parameters
    ----------
    G : graph instance
        topology-defining graph
    feature: array of shape (G.V,dim_feature)
        some vectorial information related to the graph vertices
    verbose : bool, optional
         If True, print diagnostic information

    Returns
    -------
    t: weightForest instance,
       that represents the dendrogram of the data

    Notes
    ----
    Hopefully a quicker version

    A euclidean distance is used in the feature space

    Caveat : only approximate
    """
    # basic check
    if feature.ndim == 1:
        feature = np.reshape(feature, (-1, 1))

    if feature.shape[0] != G.V:
        raise ValueError(
            "Incompatible dimension for the feature matrix and the graph")

    Features = [np.ones(2 * G.V), np.zeros((2 * G.V, feature.shape[1])),
                np.zeros((2 * G.V, feature.shape[1]))]
    Features[1][:G.V] = feature
    Features[2][:G.V] = feature ** 2
    n = G.V
    nbcc = G.cc().max() + 1

    # prepare a graph with twice the number of vertices
    K = _auxiliary_graph(G, Features)
    parent = np.arange(2 * n - nbcc).astype(np.int)
    height = np.zeros(2 * n - nbcc)
    linc = K.left_incidence()
    rinc = K.right_incidence()

    # iteratively merge clusters
    q = 0
    while (q < n - nbcc):
        # 1. find the lightest edges
        aux = np.zeros(2 * n)
        ape = np.nonzero(K.weights < np.inf)
        ape = np.reshape(ape, np.size(ape))
        idx = np.argsort(K.weights[ape])

        for e in range(n - nbcc - q):
            i, j = K.edges[ape[idx[e]], 0], K.edges[ape[idx[e]], 1]
            if (aux[i] == 1) or (aux[j] == 1):
                break
            aux[i] = 1
            aux[j] = 1

        emax = np.maximum(e, 1)

        for e in range(emax):
            m = ape[idx[e]]
            cost = K.weights[m]
            k = q + n
            i = K.edges[m, 0]
            j = K.edges[m, 1]
            height[k] = cost
            if verbose:
                print q, i, j, m, cost

            # 2. remove the current edge
            K.edges[m] = -1
            K.weights[m] = np.inf
            linc[i].remove(m)
            rinc[j].remove(m)

            ml = linc[j]
            if np.sum(K.edges[ml, 1] == i) > 0:
                m = ml[np.flatnonzero(K.edges[ml, 1] == i)]
                K.edges[m] = -1
                K.weights[m] = np.inf
                linc[j].remove(m)
                rinc[i].remove(m)

            # 3. merge the edges with third part edges
            parent[i] = k
            parent[j] = k
            for p in range(3):
                Features[p][k] = Features[p][i] + Features[p][j]

            linc, rinc = _remap(K, i, j, k, Features, linc, rinc)
            q += 1

    # build a tree to encode the results
    t = WeightedForest(2 * n - nbcc, parent, height)
    return t


def ward_field_segment(F, stop=-1, qmax=-1, verbose=False):
    """Agglomerative function based on a field structure

    Parameters
    ----------
    F the input field (graph+feature)
    stop: float, optional
        the stopping crterion.  if stop==-1, then no stopping criterion is used
    qmax: int, optional
        the maximum number of desired clusters (in the limit of the stopping
        criterion)
    verbose : bool, optional
        If True, print diagnostic information

    Returns
    -------
    u: array of shape (F.V)
       labelling of the graph vertices according to the criterion
    cost array of shape (F.V - 1)
         the cost of each merge step during the clustering procedure


    Notes
    -----
    See ward_quick_segment for more information

    Caveat : only approximate
    """
    u, cost = ward_quick_segment(F, F.field, stop, qmax, verbose)
    return u, cost


def ward_quick_segment(G, feature, stop=-1, qmax=1, verbose=False):
    """
    Agglomerative function based on a topology-defining graph
    and a feature matrix.

    Parameters
    ----------
    G: labs.graph.WeightedGraph instance
       the input graph (a topological graph essentially)
    feature array of shape (G.V,dim_feature)
        vectorial information related to the graph vertices
    stop1 : int or float, optional
        the stopping crterion if stop==-1, then no stopping criterion is used
    qmax : int, optional
        the maximum number of desired clusters (in the limit of the stopping
        criterion)
    verbose : bool, optional
        If True, print diagnostic information

    Returns
    -------
    u: array of shape (G.V)
       labelling of the graph vertices according to the criterion
    cost: array of shape (G.V - 1)
          the cost of each merge step during the clustering procedure

    Notes
    -----
    Hopefully a quicker version

    A euclidean distance is used in the feature space

    Caveat : only approximate
    """
    # basic check
    if feature.ndim == 1:
        feature = np.reshape(feature, (-1, 1))

    if feature.shape[0] != G.V:
        raise ValueError(
            "Incompatible dimension for the feature matrix and the graph")

    n = G.V
    if stop == - 1:
        stop = np.inf
    qmax = int(np.minimum(qmax, n - 1))
    t = ward_quick(G, feature, verbose)
    if verbose:
        t.plot()

    u1 = np.zeros(n, np.int)
    u2 = np.zeros(n, np.int)
    if stop >= 0:
        u1 = t.partition(stop)
    if qmax > 0:
        u2 = t.split(qmax)

    if u1.max() < u2.max():
        u = u2
    else:
        u = u1

    cost = t.get_height()
    cost = cost[t.isleaf() == False]
    return u, cost


def ward_segment(G, feature, stop=-1, qmax=1, verbose=False):
    """
    Agglomerative function based on a topology-defining graph
    and a feature matrix.

    Parameters
    ----------
    G : graph object
        the input graph (a topological graph essentially)
    feature : array of shape (G.V,dim_feature)
        some vectorial information related to the graph vertices
    stop : int or float, optional
        the stopping crterion.  if stop==-1, then no stopping criterion is used
    qmax : int, optional
        the maximum number of desired clusters (in the limit of the stopping
        criterion)
    verbose : bool, optional
        If True, print diagnostic information

    Returns
    -------
    u: array of shape (G.V):
       a labelling of the graph vertices according to the criterion
    cost: array of shape (G.V - 1)
          the cost of each merge step during the clustering procedure

    Notes
    -----
    A euclidean distance is used in the feature space

    Caveat : when the number of cc in G (nbcc) is greter than qmax, u contains
    nbcc values, not qmax !
    """
    # basic check
    if feature.ndim == 1:
        feature = np.reshape(feature, (-1, 1))

    if feature.shape[0] != G.V:
        raise ValueError(
            "Incompatible dimension for the feature matrix and the graph")

    # prepare a graph with twice the number of vertices
    n = G.V
    if qmax == -1:
        qmax = n - 1
    if stop == -1:
        stop = np.inf
    qmax = int(np.minimum(qmax, n - 1))

    t = ward(G, feature, verbose)
    u1 = np.zeros(n, np.int)
    u2 = np.zeros(n, np.int)
    if stop >= 0:
        u1 = t.partition(stop)
    if qmax > 0:
        u2 = t.split(qmax)

    if u1.max() < u2.max():
        u = u2
    else:
        u = u1

    cost = t.get_height()
    cost = cost[t.isleaf() == False]
    return u, cost


def ward(G, feature, verbose=False):
    """
    Agglomerative function based on a topology-defining graph
    and a feature matrix.

    Parameters
    ----------
    G : graph
        the input graph (a topological graph essentially)
    feature : array of shape (G.V,dim_feature)
        vectorial information related to the graph vertices
    verbose : bool, optional
        If True, print diagnostic information

    Returns
    --------
    t : ``WeightedForest`` instance
        structure that represents the dendrogram

    Notes
    -----
    When G has more than 1 connected component, t is no longer a tree.  This
    case is handled cleanly now
    """
    # basic check
    if feature.ndim == 1:
        feature = np.reshape(feature, (-1, 1))

    if feature.shape[0] != G.V:
        raise ValueError(
            "Incompatible dimension for the feature matrix and the graph")

    Features = [np.ones(2 * G.V), np.zeros((2 * G.V, feature.shape[1])),
                np.zeros((2 * G.V, feature.shape[1]))]
    Features[1][:G.V] = feature
    Features[2][:G.V] = feature ** 2

    # prepare a graph with twice the number of vertices
    # this graph will contain the connectivity information
    # along the merges.
    n = G.V
    nbcc = G.cc().max() + 1
    K = _auxiliary_graph(G, Features)

    # prepare some variables that are useful tp speed up the algorithm
    parent = np.arange(2 * n - nbcc).astype(np.int)
    height = np.zeros(2 * n - nbcc)
    linc = K.left_incidence()
    rinc = K.right_incidence()

    # iteratively merge clusters
    for q in range(n - nbcc):
        # 1. find the lightest edge
        m = (K.weights).argmin()
        cost = K.weights[m]
        k = q + n
        i = K.edges[m, 0]
        j = K.edges[m, 1]
        height[k] = cost
        if verbose:
            print q, i, j, m, cost

        # 2. remove the current edge
        K.edges[m] = - 1
        K.weights[m] = np.inf
        linc[i].remove(m)
        rinc[j].remove(m)

        ml = linc[j]
        if np.sum(K.edges[ml, 1] == i) > 0:
            m = ml[np.flatnonzero(K.edges[ml, 1] == i)]
            K.edges[m] = -1
            K.weights[m] = np.inf
            linc[j].remove(m)
            rinc[i].remove(m)

        # 3. merge the edges with third part edges
        parent[i] = k
        parent[j] = k
        for p in range(3):
            Features[p][k] = Features[p][i] + Features[p][j]

        linc, rinc = _remap(K, i, j, k, Features, linc, rinc)

    # build a tree to encode the results
    t = WeightedForest(2 * n - nbcc, parent, height)
    return t


#--------------------------------------------------------------------------
#----------------------- Visualization ------------------------------------
# -------------------------------------------------------------------------


def _label_(f, parent, left, labelled):
    temp = np.nonzero(parent == f)

    if np.size(temp) > 0:
        i = temp[0][np.nonzero(left[temp[0]] == 1)]
        j = temp[0][np.nonzero(left[temp[0]] == 0)]
        labelled = _label_(i, parent, left, labelled)
        labelled[f] = labelled.max() + 1
        labelled = _label_(j, parent, left, labelled)

    if labelled[f] < 0:
        labelled[f] = labelled.max() + 1

    return labelled


def _label(parent):
    # find the root
    root = np.nonzero(parent == np.arange(np.size(parent)))[0]

    # define left
    left = np.zeros(np.size(parent))
    for f in range(np.size(parent)):
        temp = np.nonzero(parent == f)
        if np.size(temp) > 0:
            left[temp[0][0]] = 1

    left[root] = .5

    # define labelled
    labelled = - np.ones(np.size(parent))

    # compute labelled
    for j in range(np.size(root)):
        labelled = _label_(root[j], parent, left, labelled)

    return labelled



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Wrapper to run setup.py using setuptools."""

from setuptools import setup 

################################################################################
# Call the setup.py script, injecting the setuptools-specific arguments.

extra_setuptools_args = dict(
                            tests_require=['nose'],
                            test_suite='nose.collector',
                            zip_safe=False,
                            )


if __name__ == '__main__':
    execfile('setup.py', dict(__name__='__main__', 
                          extra_setuptools_args=extra_setuptools_args))






# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Package containing core nipy classes.
"""
__docformat__ = 'restructuredtext'

from .volumes.volume_field import VolumeField
from .volumes.volume_img import VolumeImg
from .volumes.volume_grid  import VolumeGrid
from .transforms.transform import Transform, CompositionError
from .transforms.affine_transform import AffineTransform
from .transforms.affine_utils import apply_affine

from .converters import as_volume_img, save

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



# -*- Mode: Python -*-  

"""
Bindings for various image registration routines written in C: joint
histogram computation, cubic spline interpolation, non-rigid
transformations. 
"""

__version__ = '0.3'


# Includes
from numpy cimport (import_array, ndarray, flatiter, broadcast, 
                    PyArray_MultiIterNew, PyArray_MultiIter_DATA, 
                    PyArray_MultiIter_NEXT)


# Externals
cdef extern from "joint_histogram.h":
    void joint_histogram_import_array()
    int joint_histogram(ndarray H, unsigned int clampI, unsigned int clampJ,  
                        flatiter iterI, ndarray imJ_padded, 
                        ndarray Tvox, int interp)
    int L1_moments(double* n, double* median, double* dev, ndarray H)

cdef extern from "cubic_spline.h":
    void cubic_spline_import_array()
    void cubic_spline_transform(ndarray res, ndarray src)
    double cubic_spline_sample1d(double x, ndarray coef, 
                                 int mode) 
    double cubic_spline_sample2d(double x, double y, ndarray coef, 
                                 int mode_x, int mode_y) 
    double cubic_spline_sample3d(double x, double y, double z, ndarray coef,
                                 int mode_x, int mode_y, int mode_z) 
    double cubic_spline_sample4d(double x, double y, double z, double t, ndarray coef, 
                                 int mode_x, int mode_y, int mode_z, int mode_t)
    void cubic_spline_resample3d(ndarray im_resampled, ndarray im, 
                                 double* Tvox, int cast_integer,
                                 int mode_x, int mode_y, int mode_z)

cdef extern from "polyaffine.h": 
    void polyaffine_import_array()
    void apply_polyaffine(ndarray XYZ, ndarray Centers, ndarray Affines, ndarray Sigma)


# Initialize numpy
joint_histogram_import_array()
cubic_spline_import_array()
polyaffine_import_array()
import_array()
import numpy as np

# Globals
modes = {'zero': 0, 'nearest': 1, 'reflect': 2}


def _joint_histogram(ndarray H, flatiter iterI, ndarray imJ, ndarray Tvox, long interp):
    """
    Compute the joint histogram given a transformation trial. 
    """
    cdef:
        double *h, *tvox
        unsigned int clampI, clampJ
        int ret

    # Views
    clampI = <unsigned int>H.shape[0]
    clampJ = <unsigned int>H.shape[1]    

    # Compute joint histogram 
    ret = joint_histogram(H, clampI, clampJ, iterI, imJ, Tvox, interp)
    if not ret == 0:
        raise RuntimeError('Joint histogram failed because of incorrect input arrays.')

    return 


def _L1_moments(ndarray H):
    """
    Compute L1 moments of order 0, 1 and 2 of a one-dimensional
    histogram.
    """
    cdef:
        double n[1], median[1], dev[1]
        int ret

    ret = L1_moments(n, median, dev, H)
    if not ret == 0:
        raise RuntimeError('L1_moments failed because input array is not double.')

    return n[0], median[0], dev[0]


def _cspline_transform(ndarray x):
    c = np.zeros([x.shape[i] for i in range(x.ndim)], dtype=np.double)
    cubic_spline_transform(c, x)
    return c

cdef ndarray _reshaped_double(object in_arr, ndarray sh_arr):
    shape = [sh_arr.shape[i] for i in range(sh_arr.ndim)]
    return np.reshape(in_arr, shape).astype(np.double)

def _cspline_sample1d(ndarray R, ndarray C, X=0, mode='zero'):
    cdef double *r, *x
    cdef broadcast multi
    Xa = _reshaped_double(X, R) 
    multi = PyArray_MultiIterNew(2, <void*>R, <void*>Xa)
    while(multi.index < multi.size):
        r = <double*>PyArray_MultiIter_DATA(multi, 0)
        x = <double*>PyArray_MultiIter_DATA(multi, 1)
        r[0] = cubic_spline_sample1d(x[0], C, modes[mode])
        PyArray_MultiIter_NEXT(multi)
    return R

def _cspline_sample2d(ndarray R, ndarray C, X=0, Y=0, 
                      mx='zero', my='zero'):
    cdef double *r, *x, *y
    cdef broadcast multi
    Xa = _reshaped_double(X, R)
    Ya = _reshaped_double(Y, R)
    multi = PyArray_MultiIterNew(3, <void*>R, <void*>Xa, <void*>Ya)
    while(multi.index < multi.size):
        r = <double*>PyArray_MultiIter_DATA(multi, 0)
        x = <double*>PyArray_MultiIter_DATA(multi, 1)
        y = <double*>PyArray_MultiIter_DATA(multi, 2)
        r[0] = cubic_spline_sample2d(x[0], y[0], C, modes[mx], modes[my])
        PyArray_MultiIter_NEXT(multi)
    return R

def _cspline_sample3d(ndarray R, ndarray C, X=0, Y=0, Z=0, 
                      mx='zero', my='zero', mz='zero'):
    cdef double *r, *x, *y, *z
    cdef broadcast multi
    Xa = _reshaped_double(X, R)
    Ya = _reshaped_double(Y, R)
    Za = _reshaped_double(Z, R)
    multi = PyArray_MultiIterNew(4, <void*>R, <void*>Xa, <void*>Ya, <void*>Za)
    while(multi.index < multi.size):
        r = <double*>PyArray_MultiIter_DATA(multi, 0)
        x = <double*>PyArray_MultiIter_DATA(multi, 1)
        y = <double*>PyArray_MultiIter_DATA(multi, 2)
        z = <double*>PyArray_MultiIter_DATA(multi, 3)
        r[0] = cubic_spline_sample3d(x[0], y[0], z[0], C, modes[mx], modes[my], modes[mz])
        PyArray_MultiIter_NEXT(multi)
    return R


def _cspline_sample4d(ndarray R, ndarray C, X=0, Y=0, Z=0, T=0, 
                      mx='zero', my='zero', mz='zero', mt='zero'):
    """
    In-place cubic spline sampling. R.dtype must be 'double'. 
    """
    cdef double *r, *x, *y, *z, *t
    cdef broadcast multi
    Xa = _reshaped_double(X, R)
    Ya = _reshaped_double(Y, R)
    Za = _reshaped_double(Z, R)
    Ta = _reshaped_double(T, R)
    multi = PyArray_MultiIterNew(5, <void*>R, <void*>Xa, <void*>Ya, <void*>Za, <void*>Ta)
    while(multi.index < multi.size):
        r = <double*>PyArray_MultiIter_DATA(multi, 0)
        x = <double*>PyArray_MultiIter_DATA(multi, 1)
        y = <double*>PyArray_MultiIter_DATA(multi, 2)
        z = <double*>PyArray_MultiIter_DATA(multi, 3)
        t = <double*>PyArray_MultiIter_DATA(multi, 4)
        r[0] = cubic_spline_sample4d(x[0], y[0], z[0], t[0], C, modes[mx], modes[my], modes[mz], modes[mt])
        PyArray_MultiIter_NEXT(multi)
    return R


def _cspline_resample3d(ndarray im, dims, ndarray Tvox, dtype=None,
                        mx='zero', my='zero', mz='zero'):
    """
    Perform cubic spline resampling of a 3d input image `im` into a
    grid with shape `dims` according to an affine transform
    represented by a 4x4 matrix `Tvox` that assumes voxel
    coordinates. Boundary conditions on each axis are determined by
    the keyword arguments `mx`, `my` and `mz`, respectively. Possible
    choices are:

    'zero': assume zero intensity outside the target grid
    'nearest': extrapolate intensity by the closest grid point along the axis
    'reflect': extrapolate intensity by mirroring the input image along the axis

    Note that `Tvox` will be re-ordered in C convention if needed.
    """
    cdef double *tvox
    cdef int cast_integer

    # Create output array
    if dtype == None:
        dtype = im.dtype
    im_resampled = np.zeros(tuple(dims), dtype=dtype)

    # Ensure that the Tvox array is C-contiguous (required by the
    # underlying C routine)
    Tvox = np.asarray(Tvox, dtype='double', order='C')
    tvox = <double*>Tvox.data

    # Actual resampling 
    if dtype.kind == 'i':
        cast_integer = 1
    elif dtype.kind == 'u':
        cast_integer = 2
    else:
        cast_integer = 0
    cubic_spline_resample3d(im_resampled, im, tvox, cast_integer,
                            modes[mx], modes[my], modes[mz])

    return im_resampled


def check_array(ndarray x, int dim, int exp_dim, xname): 
    if not x.flags['C_CONTIGUOUS'] or not x.dtype=='double':
        raise ValueError('%s array should be double C-contiguous' % xname)
    if not dim == exp_dim: 
        raise ValueError('%s has size %d in last dimension, %d expected' % (xname, dim, exp_dim))

def _apply_polyaffine(ndarray xyz, ndarray centers, ndarray affines, ndarray sigma): 

    check_array(xyz, xyz.shape[1], 3, 'xyz') 
    check_array(centers, centers.shape[1], 3, 'centers')
    check_array(affines, affines.shape[1], 12, 'affines')
    check_array(sigma, sigma.size, 3, 'sigma')
    if not centers.shape[0] == affines.shape[0]: 
        raise ValueError('centers and affines arrays should have same shape[0]')

    apply_polyaffine(xyz, centers, affines, sigma) 



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Plot of the canonical Glover HRF
"""

import numpy as np

from nipy.modalities.fmri import hrf, utils

import matplotlib.pyplot as plt

# hrf.glover is a symbolic function; get a function of time to work on arrays
hrf_func = utils.lambdify_t(hrf.glover(utils.T))

t = np.linspace(0,25,200)
plt.plot(t, hrf_func(t))
a=plt.gca()
a.set_xlabel(r'$t$')
a.set_ylabel(r'$h_{can}(t)$')



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This script requires the nipy-data package to run. It is an example of
inter-subject affine registration using two MR-T1 images from the
sulcal 2000 database acquired at CEA, SHFJ, Orsay, France. The source
is 'ammon' and the target is 'anubis'. Running it will result in a
resampled ammon image being created in the current directory.
"""
from __future__ import print_function # Python 2/3 compatibility

from optparse import OptionParser
import time

import numpy as np

from nipy.algorithms.registration import HistogramRegistration, resample
from nipy.utils import example_data
from nipy import load_image, save_image

print('Scanning data directory...')

# Input images are provided with the nipy-data package
source = 'ammon'
target = 'anubis'
source_file = example_data.get_filename('neurospin', 'sulcal2000',
                                        'nobias_' + source + '.nii.gz')
target_file = example_data.get_filename('neurospin', 'sulcal2000',
                                        'nobias_' + target + '.nii.gz')

# Parse arguments
parser = OptionParser(description=__doc__)

doc_similarity = 'similarity measure: cc (correlation coefficient), \
cr (correlation ratio), crl1 (correlation ratio in L1 norm), \
mi (mutual information), nmi (normalized mutual information), \
pmi (Parzen mutual information), dpmi (discrete Parzen mutual \
information). Default is crl1.'

doc_interp = 'interpolation method: tri (trilinear), pv (partial volume), \
rand (random). Default is pv.'

doc_optimizer = 'optimization method: simplex, powell, steepest, cg, bfgs. \
Default is powell.'

parser.add_option('-s', '--similarity', dest='similarity',
                  help=doc_similarity)
parser.add_option('-i', '--interp', dest='interp',
                  help=doc_interp)
parser.add_option('-o', '--optimizer', dest='optimizer',
                  help=doc_optimizer)
opts, args = parser.parse_args()


# Optional arguments
similarity = 'crl1'
interp = 'pv'
optimizer = 'powell'
if not opts.similarity == None:
    similarity = opts.similarity
if not opts.interp == None:
    interp = opts.interp
if not opts.optimizer == None:
    optimizer = opts.optimizer

# Print messages
print('Source brain: %s' % source)
print('Target brain: %s' % target)
print('Similarity measure: %s' % similarity)
print('Optimizer: %s' % optimizer)

# Get data
print('Fetching image data...')
I = load_image(source_file)
J = load_image(target_file)

# Perform affine registration
# The output is an array-like object such that
# np.asarray(T) is a customary 4x4 matrix
print('Setting up registration...')
tic = time.time()
R = HistogramRegistration(I, J, similarity=similarity, interp=interp)
T = R.optimize('affine', optimizer=optimizer)
toc = time.time()
print('  Registration time: %f sec' % (toc - tic))

# Resample source image
print('Resampling source image...')
tic = time.time()
#It = resample2(I, J.coordmap, T.inv(), J.shape)
It = resample(I, T.inv(), reference=J)
toc = time.time()
print('  Resampling time: %f sec' % (toc - tic))

# Save resampled source
outroot = source + '_TO_' + target
outimg = outroot + '.nii.gz'
print ('Saving resampled source in: %s' % outimg)
save_image(It, outimg)

# Save transformation matrix
outparams = outroot + '.npy'
np.save(outparams, np.asarray(T))



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import print_function # Python 2/3 compatibility
__doc__ = """
Example of activation image visualization with nipy.labs visualization tools

Needs *example data* package.

Needs matplotlib
"""
print(__doc__)

import os.path

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise RuntimeError("This script needs the matplotlib library")

from nibabel import load

from nipy.labs import viz
from nipy.utils import example_data

# Local import
from get_data_light import get_second_level_dataset

# get the data
data_dir = get_second_level_dataset()

# First example, with a anatomical template
img = load(os.path.join(data_dir, 'spmT_0029.nii.gz'))
data = img.get_data()
affine = img.get_affine()

viz.plot_map(data, affine, cut_coords=(-52, 10, 22),
                        threshold=2.0, cmap=viz.cm.cold_hot)
plt.savefig('ortho_view.png')

# Second example, with a given anatomical image slicing in the Z direction
try:
    anat_img = load(example_data.get_filename('neurospin', 'sulcal2000',
                                              'nobias_anubis.nii.gz'))
    anat = anat_img.get_data()
    anat_affine = anat_img.get_affine()
except OSError as e:
    # File does not exist: the data package is not installed
    print(e)
    anat = None
    anat_affine = None

viz.plot_map(data, affine, anat=anat, anat_affine=anat_affine,
             slicer='z', threshold=2, cmap=viz.cm.cold_hot, black_bg=True)
plt.savefig('z_view.png')

viz.plot_map(data, affine, anat=anat, anat_affine=anat_affine,
             slicer='x', threshold=2, cmap=viz.cm.cold_hot, black_bg=True)
plt.savefig('x_view.png')

viz.plot_map(data, affine, anat=anat, anat_affine=anat_affine,
             slicer='y', threshold=2, cmap=viz.cm.cold_hot, black_bg=True)
plt.savefig('y_view.png')

plt.show()



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Testing VolumeGrid interface.
"""

import nose
import copy

import numpy as np

# Local imports
from ..volume_grid import VolumeGrid
from ..volume_img import VolumeImg
from ...transforms.transform import Transform, CompositionError

def mapping(x, y, z):
    return 2*x, y, 0.5*z

def inverse_mapping(x, y, z):
    return 0.5*x, y, 2*z

def id(x, y, z):
    return x, y, z

################################################################################
# Tests
def test_constructor():
    yield np.testing.assert_raises, ValueError, VolumeGrid, None, \
        None, {}, 'e'


def test_volume_grid():
    """ Sanity testing of the VolumeGrid class.
    """
    transform = Transform('voxels', 'world', mapping)
    img = VolumeGrid(data=np.random.random((10, 10, 10)),
                    transform=transform,
                    )
    # Test that the repr doesn't raise an error
    yield repr, img

    # We cannot calculate the values in the world, because the transform 
    # is not invertible.
    
    yield np.testing.assert_raises, ValueError, \
                        img.values_in_world, 0, 0, 0
    yield np.testing.assert_raises, ValueError, \
        img.as_volume_img

    yield nose.tools.assert_equal, img, copy.copy(img)


def test_trivial_grid():
    """ Test resampling for an grid embedded in world space with an
        identity transform. 
    """
    N = 10
    identity = Transform('voxels', 'world', id, id)
    data = np.random.random((N, N, N))
    img = VolumeGrid(data=data,
                    transform=identity,
                    )
    x, y, z = np.random.random_integers(N, size=(3, 10)) - 1
    data_ = img.values_in_world(x, y, z)
    # Check that passing in arrays with different shapes raises an error
    yield np.testing.assert_raises, ValueError, \
        img.values_in_world, x, y, z[:-1]
    # Check that passing in wrong interpolation keyword raises an error
    yield np.testing.assert_raises, ValueError, \
                        img.values_in_world, 0, 0, 0, 'e'
    yield np.testing.assert_almost_equal, data[x, y, z], data_


def test_transformation():
    """ Test transforming images.
    """
    N = 10
    v2w_mapping = Transform('voxels', 'world1', mapping, 
                            inverse_mapping)
    identity  = Transform('world1', 'world2', id, id) 
    data = np.random.random((N, N, N))
    img1 = VolumeGrid(data=data,
                     transform=v2w_mapping,
                     )
    img2 = img1.composed_with_transform(identity)
    
    yield nose.tools.assert_equal, img2.world_space, 'world2'

    x, y, z = N*np.random.random(size=(3, 10))
    yield np.testing.assert_almost_equal, img1.values_in_world(x, y, z), \
        img2.values_in_world(x, y, z)

    yield nose.tools.assert_raises, CompositionError, \
            img1.composed_with_transform, identity.get_inverse()

    yield nose.tools.assert_raises, CompositionError, img1.resampled_to_img, \
            img2
    
    # Resample an image on itself: it shouldn't change much:
    img  = img1.resampled_to_img(img1)
    yield np.testing.assert_almost_equal, data, img.get_data()

    # Check that if I 'resampled_to_img' on an VolumeImg, I get an
    # VolumeImg, and vice versa 
    volume_image = VolumeImg(data, np.eye(4), 'world')
    identity  = Transform('voxels', 'world', id, id) 
    image = VolumeGrid(data, identity)
    image2 = image.resampled_to_img(volume_image)
    yield nose.tools.assert_true, isinstance(image2, VolumeImg)
    volume_image2 = volume_image.resampled_to_img(image)
    yield nose.tools.assert_true, isinstance(image2, VolumeGrid)
    # Check that the data are all the same: we have been playing only
    # with identity mappings
    yield np.testing.assert_array_equal, volume_image2.get_data(), \
            image2.get_data()


def test_as_volume_image():
    """ Test casting VolumeGrid to VolumeImg
    """
    N = 10
    v2w_mapping  = Transform('voxels', 'world2', id, id) 
    data = np.random.random((N, N, N))
    img1 = VolumeGrid(data=data,
                     transform=v2w_mapping,
                     )
    img2 = img1.as_volume_img()

    # Check that passing in the wrong shape raises an error
    yield nose.tools.assert_raises, ValueError, img1.as_volume_img, None, \
            (10, 10)




# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:

import numpy as np

from .... import load_image
from ....testing import anatfile
from ..histogram_registration import HistogramRegistration

from numpy.testing import assert_array_almost_equal

anat_img = load_image(anatfile)

def test_registers():
    # Test registration to self returns identity
    for cost, interp, affine_type in (('cc', 'pv', 'rigid'),
                                      ('cc', 'tri', 'rigid'),
                                      ('cc', 'rand', 'rigid'),
                                      ('cc', 'pv', 'similarity'),
                                      ('cc', 'pv', 'affine'),
                                      ('cr', 'pv', 'rigid'),
                                      ('cr', 'pv', 'rigid'),
                                      ('crl1', 'pv', 'rigid'),
                                      ('mi', 'pv', 'rigid'),
                                      ('nmi', 'pv', 'rigid'),
                                     ):
        R = HistogramRegistration(anat_img, anat_img,
                                  similarity=cost,
                                  interp=interp)
        R.subsample([2,2,2])
        affine = R.optimize(affine_type)
        yield assert_array_almost_equal, affine.as_affine(), np.eye(4), 2



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Installation script for nipy examplepkg package '''
import os
from os.path import join as pjoin

from distutils.core import setup

try:
    import ConfigParser as cfg # Python 2
except ImportError:
    import configparser as cfg # Python 3

# The directory under --prefix, under which to store files
OUTPUT_BASE = pjoin('share', 'nipy', 'nipy')

# The directory in this directory to be copied into OUTPUT_BASE
# such that <prefix>/<OUPUT_BASE>/<PKG_BASE> will exist
PKG_BASE = 'examplepkg'

DATA_FILES = []

for dirpath, dirnames, filenames in os.walk(PKG_BASE):
    files = [pjoin(dirpath, filename) for filename in filenames]
    DATA_FILES.append((pjoin(OUTPUT_BASE, dirpath), files))

config = cfg.SafeConfigParser()
config.read(pjoin(PKG_BASE, 'config.ini'))

setup(
    name = 'nipy-' + PKG_BASE,
    version = config.get('DEFAULT', 'version'),
    description='NIPY %s data package' % PKG_BASE,
    author='The NIPY team',
    url='http://neuroimaging.scipy.org',
    author_email='nipy-devel@neuroimaging.scipy.org',
    data_files = DATA_FILES,
    )




from ._registration import _L1_moments

import numpy as np
from scipy.ndimage import gaussian_filter

TINY = float(np.finfo(np.double).tiny)
SIGMA_FACTOR = 0.05


def nonzero(x):
    """
    Force strictly positive values.
    """
    return np.maximum(x, TINY)


def dist2loss(dist, margI=None, margJ=None):
    L = dist
    LT = L.T
    if margI == None:
        margI = L.sum(0)
    if margJ == None:
        margJ = L.sum(1)
    L /= nonzero(margI)
    LT /= nonzero(margJ)
    return -np.log(nonzero(L))


class SimilarityMeasure(object):

    def __init__(self, shape, **kwargs):
        self.shape = shape
        self.J, self.I = np.indices(shape)
        for key in kwargs.keys():
            self.__setattr__(key, kwargs[key])

    def loss(self, H):
        return np.zeros(H.shape)

    def npoints(self, H):
        return H.sum()

    def overall_loss(self, H):
        return np.sum(H * self.loss(H))

    def averaged_loss(self, H):
        return np.sum(H * self.loss(H)) / nonzero(self.npoints(H))

    def __call__(self, H):
        return -self.averaged_loss(H)


class SupervisedLikelihoodRatio(SimilarityMeasure):

    def loss(self, H):
        if not hasattr(self, 'L'):
            self.L = dist2loss(self.dist)
        return self.L


class MutualInformation(SimilarityMeasure):

    def loss(self, H):
        return dist2loss(H / nonzero(self.npoints(H)))


class ParzenMutualInformation(SimilarityMeasure):

    def loss(self, H):
        if not hasattr(self, 'sigma'):
            self.sigma = SIGMA_FACTOR * np.array(H.shape)
        npts = nonzero(self.npoints(H))
        Hs = H / npts
        gaussian_filter(Hs, sigma=self.sigma, mode='constant', output=Hs)
        return dist2loss(Hs)


class DiscreteParzenMutualInformation(SimilarityMeasure):

    def loss(self, H):
        if not hasattr(self, 'sigma'):
            self.sigma = SIGMA_FACTOR * np.array(H.shape)
        Hs = gaussian_filter(H, sigma=self.sigma, mode='constant')
        Hs /= nonzero(Hs.sum())
        return dist2loss(Hs)


class NormalizedMutualInformation(SimilarityMeasure):
    """
    NMI = 2*(1 - H(I,J)/[H(I)+H(J)])
        = 2*MI/[H(I)+H(J)])
    """
    def loss(self, H):
        L = H / nonzero(self.npoints(H))
        lI = L.sum(0)
        lJ = L.sum(1)
        self.hI = lI
        self.hJ = lJ
        return -np.log(nonzero(L))

    def __call__(self, H):
        HIJ = self.averaged_loss(H)
        HI = -np.sum(self.hI * np.log(nonzero(self.hI)))
        HJ = -np.sum(self.hJ * np.log(nonzero(self.hJ)))
        return 2 * (1 - HIJ / nonzero(HI + HJ))


class CorrelationCoefficient(SimilarityMeasure):

    def loss(self, H):
        rho2 = self(H)
        I = (self.I - self.mI) / np.sqrt(nonzero(self.vI))
        J = (self.J - self.mJ) / np.sqrt(nonzero(self.vJ))
        L = rho2 * I ** 2 + rho2 * J ** 2 - 2 * self.rho * I * J
        tmp = nonzero(1. - rho2)
        L *= .5 / tmp
        L += .5 * np.log(tmp)
        return L

    def averaged_loss(self, H):
        return .5 * np.log(nonzero(1 - self(H)))

    def __call__(self, H):
        npts = nonzero(self.npoints(H))
        self.mI = np.sum(H * self.I) / npts
        self.mJ = np.sum(H * self.J) / npts
        self.vI = np.sum(H * (self.I) ** 2) / npts - self.mI ** 2
        self.vJ = np.sum(H * (self.J) ** 2) / npts - self.mJ ** 2
        self.cIJ = np.sum(H * self.J * self.I) / npts - self.mI * self.mJ
        self.rho = self.cIJ / nonzero(np.sqrt(self.vI * self.vJ))
        return self.rho ** 2


class CorrelationRatio(SimilarityMeasure):

    def loss(self, H):
        print('Sorry, not implemented yet...')
        return

    def averaged_loss(self, H):
        return .5 * np.log(nonzero(1. - self(H)))

    def __call__(self, H):
        self.npts_J = np.sum(H, 1)
        tmp = nonzero(self.npts_J)
        self.mI_J = np.sum(H * self.I, 1) / tmp
        self.vI_J = np.sum(H * (self.I) ** 2, 1) / tmp - self.mI_J ** 2
        self.npts = np.sum(self.npts_J)
        tmp = nonzero(self.npts)
        hI = np.sum(H, 0)
        hJ = np.sum(H, 1)
        self.mI = np.sum(hI * self.I[0, :]) / tmp
        self.vI = np.sum(hI * self.I[0, :] ** 2) / tmp - self.mI ** 2
        mean_vI_J = np.sum(hJ * self.vI_J) / tmp
        return 1. - mean_vI_J / nonzero(self.vI)


class CorrelationRatioL1(SimilarityMeasure):     

    def loss(self, H):
        print('Sorry, not implemented yet...')
        return

    def averaged_loss(self, H):
        return np.log(nonzero(1. - self(H)))

    def __call__(self, H):
        tmp = np.array([_L1_moments(H[j, :]) for j in range(H.shape[0])])
        self.npts_J, self.mI_J, self.sI_J = tmp[:, 0], tmp[:, 1], tmp[:, 2]
        hI = np.sum(H, 0)
        hJ = np.sum(H, 1)
        self.npts, self.mI, self.sI = _L1_moments(hI)
        mean_sI_J = np.sum(hJ * self.sI_J) / nonzero(self.npts)
        return 1. - mean_sI_J / nonzero(self.sI)


similarity_measures = {
    'slr': SupervisedLikelihoodRatio,
    'mi': MutualInformation,
    'nmi': NormalizedMutualInformation,
    'pmi': ParzenMutualInformation,
    'dpmi': DiscreteParzenMutualInformation,
    'cc': CorrelationCoefficient,
    'cr': CorrelationRatio,
    'crl1': CorrelationRatioL1}



#!/usr/bin/env python
import numpy as np

from ..field import (field_from_coo_matrix_and_data,
                     field_from_graph_and_data)
from ..graph import wgraph_from_3d_grid

from nose.tools import assert_true, assert_equal

from numpy.testing import TestCase, assert_array_equal


def basic_field(nx=10, ny=10, nz=10):
    xyz = np.reshape(np.indices((nx, ny, nz)), (3, nx * ny * nz)).T
    data = np.sum(xyz, 1).astype('d')
    F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 26), data)
    return F


def basic_field_random(nx=10, ny=10, nz=1):
    import numpy.random as nr
    xyz = np.reshape(np.indices((nx, ny, nz)), (3, nx * ny * nz)).T
    data = 0.5 * nr.randn(nx * ny * nz, 1) + np.sum(xyz, 1).astype('d')
    F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 26), data)
    return F


def basic_field_2(nx=10, ny=10, nz=10):
    xyz = np.reshape(np.indices((nx, ny, nz)), (3, nx * ny * nz)).T
    toto = xyz - np.array([5, 5, 5])
    data = np.sum(toto ** 2, 1)
    F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 26), data)
    return F


def basic_field_3(nx=10, ny=10, nz=10):
    xyz = np.reshape(np.indices((nx, ny, nz)), (3, nx * ny * nz)).T
    toto = xyz - np.array([5, 5, 5])
    data = np.abs(np.sum(toto ** 2, 1) - 11 )
    F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 26), data)
    return F


def basic_graph(nx=10, ny=10, nz=10):
    xyz = np.reshape(np.indices((nx, ny, nz)), (3, nx * ny * nz)).T
    data = np.zeros(xyz.shape[0])
    F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 26), data)
    return F


class test_Field(TestCase):

    def test_max_1(self):
        F  = basic_field()
        F.field[555] = 30
        depth = F.local_maxima()
        dep = np.zeros(1000, np.int)
        dep[555] = 5
        dep[999] = 3
        assert_true(sum(np.absolute(dep-depth))<1.e-7)

    def test_max_2(self):
        F  = basic_field()
        F.field[555] = 28
        idx,depth = F.get_local_maxima()
        self.assert_(len(idx) == 2)
        self.assert_(np.alltrue( idx == (555, 999)))
        self.assert_(np.alltrue( depth == (5, 3)))

    def test_max_3(self):
        F  = basic_field()
        F.field[555] = 27
        idx, depth = F.get_local_maxima()
        assert_true(np.size(idx) == 2)
        assert_true(idx[0] == 555)
        assert_true(idx[1] == 999)
        assert_true(depth[0] == 5)
        assert_true(depth[1] == 5)

    def test_max_4(self):
        F  = basic_field()
        F.field[555] = 28
        idx, depth = F.get_local_maxima(0, 27.5)
        assert_true(np.size(idx) == 1)
        assert_true(idx[0] == 555)
        assert_true(depth[0] == 1)

    def test_smooth_1(self):
        G  = basic_graph()
        field = np.zeros((1000,1))
        field[555,0] = 1
        G.set_field(field)
        G.diffusion()
        sfield = G.get_field()
        OK1 = (sfield[555]==0)
        OK2 = (sfield[554]==1)
        OK3 = (np.absolute(sfield[566]-np.sqrt(2))<1.e-7)
        OK4 = (np.absolute(sfield[446]-np.sqrt(3))<1.e-7)
        OK = OK1 & OK2 & OK3 & OK4
        self.assert_(OK)

    def test_smooth_2(self):
        G  = basic_graph()
        field = np.zeros((1000,1))
        field[555,0] = 1
        G.set_field(field)
        G.diffusion(1)
        sfield = G.get_field()
        OK1 = (sfield[555]==0)
        OK2 = (sfield[554]==1)
        OK3 = (np.absolute(sfield[566]-np.sqrt(2))<1.e-7)
        OK4 = (np.absolute(sfield[446]-np.sqrt(3))<1.e-7)
        OK = OK1 & OK2 & OK3 & OK4
        self.assert_(OK)

    def test_dilation(self):
        F  = basic_field()
        F.field[555] = 30
        F.field[664] = 0
        F.dilation(2)
        assert_true(F.field[737] == 30)
        assert_true(F.field[0] == 6)
        assert_true(F.field[999] == 27)
        assert_true(F.field[664] == 30)

    def test_dilation2(self):
        # test equality of cython and python versions
        F  = basic_field()
        F.field[555] = 30
        F.field[664] = 0
        h = F.copy()
        h.dilation(2)
        g = F.copy()
        g.dilation(2, False)
        assert_array_equal(h.field, g.field)

    def test_erosion(self):
        F  = basic_field()
        F.field[555] = 30
        F.field[664] = 0
        F.erosion(2)
        field = F.get_field()
        assert_true(field[737] == 11)
        assert_true(field[0] == 0)
        assert_true(field[999] == 21)
        assert_true(field[664] == 0)

    def test_opening(self):
        F  = basic_field()
        F.field[555] = 30
        F.field[664] = 0
        F.opening(2)
        field = F.get_field()
        assert_true(field[737] == 17)
        assert_true(field[0] == 0)
        assert_true(field[999] == 21)
        assert_true(field[555] == 16)

    def test_closing(self):
        F  = basic_field()
        F.field[555] = 30
        F.field[664] = 0
        F.closing(2)
        field = F.get_field()
        assert_true(field[737] == 17)
        assert_true(field[0] == 6)
        assert_true(field[999] == 27)
        assert_true(field[555] == 30)

    def test_watershed_1(self):
        F = basic_field()
        F.field[555] = 28
        F.field[664] = 0
        idx, label = F.custom_watershed()
        assert_equal(np.size(idx), 2)
        assert_equal(tuple(idx), (555, 999))
        assert_equal((label[776], label[666], label[123]), (1, 0, 0))

    def test_watershed_4(self):
        F = basic_field_3()
        idx, label = F.custom_watershed()
        assert_true(np.size(idx) == 9)
        assert_true(np.unique(
            [label[555], label[0], label[9], label[90], label[99], label[900],
             label[909], label[990], label[999]])
            .size == 9)

    def test_watershed_2(self):
        F = basic_field_2()
        F.field[555] = 10
        F.field[664] = 0
        idx, label = F.custom_watershed()
        assert_true(np.size(idx)==9)

    def test_watershed_3(self):
        F  = basic_field_2()
        F.field[555] = 10
        F.field[664] = 0
        idx, label = F.custom_watershed(0,11)
        assert_true(np.size(idx)==8)

    def test_bifurcations_1(self):
        F = basic_field()
        idx, parent,label = F.threshold_bifurcations()
        assert_true(idx == 999)
        assert_true(parent == 0)

    def test_bifurcations_2(self):
        F = basic_field_2()
        idx, parent, label = F.threshold_bifurcations()
        assert_true(np.size(idx) == 15)

    def test_geodesic_kmeans(self, nbseeds=3):
        # Test the geodisc k-means algorithm
        import numpy.random as nr
        F = basic_field_random(5, 5, 1)
        seeds = np.argsort(nr.rand(F.V))[:nbseeds]
        seeds, label, inertia = F.geodesic_kmeans(seeds)
        for i in range(nbseeds):
            assert_true(label[seeds[i]] == i)
        print np.unique(label), np.arange(nbseeds)
        assert_true(np.array([i in np.unique(label)
                              for i in np.arange(nbseeds)]).all())

    def test_constrained_voronoi(self, nbseeds=3):
        # Test the geodisc k-means algorithm
        import numpy.random as nr
        F = basic_field_random()
        seeds = np.argsort(nr.rand(F.V))[:nbseeds]
        label = F.constrained_voronoi(seeds)
        for i in range(nbseeds):
            assert_true(label[seeds[i]] == i)
        assert_true(np.array([i in np.unique(label)
                              for i in np.arange(nbseeds)]).all())

    def test_constrained_voronoi_2(self, nbseeds=3):
        # Test the geodisc k-means algorithm
        xyz, x = np.zeros((30, 3)), np.arange(30)
        xyz[:, 0] = x
        y = np.array((x // 10), np.float)
        F = field_from_graph_and_data(wgraph_from_3d_grid(xyz, 6),  y)
        seeds = np.array([1, 18, 25])
        label = F.constrained_voronoi(seeds)
        assert_array_equal(label, x // 10)

    def test_subfield(self):
        import numpy.random as nr
        F = basic_field_random()
        valid = nr.rand(F.V)>0.1
        sf = F.subfield(valid)
        self.assert_(sf.V==np.sum(valid))

    def test_subfield2(self):
        F = basic_field_random()
        valid = np.zeros(F.V)
        sf = F.subfield(valid)
        self.assert_(sf==None)

    def test_ward1(self):
        F = basic_field_random()
        Lab, J = F.ward(10)
        self.assert_(Lab.max()==9)

    def test_ward2(self):
        F = basic_field_random()
        Lab, J1 = F.ward(5)
        Lab, J2 = F.ward(10)
        self.assert_(J1>J2)

    def test_field_from_coo_matrix(self):
        import scipy.sparse as sps
        V = 10
        a = np.random.rand(V, V)>.9
        fi = field_from_coo_matrix_and_data(sps.coo_matrix(a), a)
        print fi.E , a.sum()
        self.assert_(fi.E==a.sum())

if __name__ == '__main__':
    import nose
    nose.run(argv=['', __file__])



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
''' Script to do very basic conversion of TRAC to reST format '''

import sys
import re

rst_section_levels = ['*', '=', '-', '~', '^']
heading_re = re.compile('^(=+) +([^=]+) +(=+)')
enumerate_re = re.compile(r'^ +(\d+\)) +(.*)')
link_re = re.compile(r'\[+ *(http://[\w.\-~/]+) +(.+?) *\]+')
italic_re = re.compile(r"'''(.+?)'''")
bold_re = re.compile(r"''(.+?)''")
inpre_re = re.compile(r"{{{(.+?)}}}")
outprestart_re = re.compile(r"^ *{{{ *$")
outprestop_re = re.compile(r"^ *}}} *$")
define_re = re.compile("^([^ ])(.*?)::")


class state_object(object):
    def __init__(self, lines, indent=''):
        self.lines = lines
        self.indent = indent

    def add_line(self, line):
        self.lines.append(self.indent+line)

    def __call__(self, line):
        raise NotImplementedError

    def __iter__(self):
        for line in self.lines:
            yield line


class preformatted_state(state_object):
    ''' State function for within preformatted blocks '''
    def __call__(self, line):
        if outprestop_re.match(line):
            self.add_line('')
            return standard_state(self.lines)
        self.add_line('   ' + line)
        return self


def std_subs(line):
    ''' standard in-line substitutions '''
    line = link_re.sub(r'`\2 <\1>`_', line)
    line = italic_re.sub(r'*\1*', line)
    line = bold_re.sub(r'**\1**', line)
    line = inpre_re.sub(r'``\1``', line)
    return line


class standard_state(state_object):
    def __call__(self, line):
        ''' State function for within normal text '''
        # beginning preformat block?
        if outprestart_re.match(line):
            self.add_line('::')
            self.add_line('')
            return preformatted_state(self.lines)
        # Heading
        hmatch = heading_re.match(line)
        if hmatch:
            eq1, heading, eq2 = hmatch.groups()
            if len(eq1) == len(eq2):
                self.add_line(heading)
                self.add_line(rst_section_levels[len(eq1)] * len(heading))
                return self
        if line.startswith(' *'):
            line = line[1:]
        line = enumerate_re.sub(r'#. \2', line)
        line = define_re.sub(r'\1\2', line)
        line = std_subs(line)
        self.add_line(line)
        return self


def trac2rst(linesource):
    ''' Process trac line source 

    A simple finite state machine

    >>> lines = ['Hello', '= Heading1 =', '=Heading2=', '== Heading 3 ==']
    >>> trac2rst(lines)
    ['Hello', 'Heading1', '========', '=Heading2=', 'Heading 3', '---------']
    >>> trac2rst([' * bullet point'])
    ['* bullet point']
    >>> trac2rst([' 33 not numbered'])
    [' 33 not numbered']
    >>> trac2rst([' 33) numbered'])
    ['#. numbered']
    >>> trac2rst(['some text then [http://www.python.org/doc a link], then text'])
    ['some text then `a link <http://www.python.org/doc>`_, then text']
    >>> line = 'text [http://www.python.org python] ' + \
               'text [http://www.scipy.org scipy] '
    >>> trac2rst([line])
    ['text `python <http://www.python.org>`_ text `scipy <http://www.scipy.org>`_']
    >>> # The next line conceals the triple quotes from the docstring parser
    >>> trac2rst([r"Some %sitalic%s text, then %smore%s" % (("'"*3,) * 4)])
    ['Some *italic* text, then *more*']
    >>> trac2rst([r"Some ''bold'' text, then ''more''"])
    ['Some **bold** text, then **more**']
    >>> # inline preformatted text
    >>> trac2rst(['here is some {{{preformatted text}}} (end)'])
    ['here is some ``preformatted text`` (end)']
    >>> # multiline preformatted
    >>> trac2rst(['','{{{','= preformatted =', ' * text', '}}}'])
    ['', '::', '', '   = preformatted =', '    * text', '']
    >>> # define
    >>> trac2rst(['a definition::', '  some explanation'])
    ['a definition', '  some explanation']
    '''
    processor = standard_state([])
    for line in linesource:
        line = line.rstrip()
        processor = processor(line)
    return processor.lines


if __name__ == '__main__':
    try:
        infile = sys.argv[1]
    except IndexError:
        raise OSError('Need input file')
    lines = trac2rst(file(infile))
    print '\n'.join(lines)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np
from numpy.linalg import svd

from reml import orth

def _trace(x):
    """
    Trace of a square 2d array.

    Does not check
    shape of x to ensure it's square.
    """
    return np.diag(x).sum()

def _frobenius(A, B):
    """
    Frobenius inner product of A and B: Trace(A'B)

    Does not check
    shape of x to ensure it's square.
    """
    return (A * B).sum()

def trRV(X=None, V=None):
    """
    If V is None it defaults to identity.

    If X is None, it defaults to the 0-dimensional subspace,
    i.e. R is the identity.

    >>> import numpy as np
    >>> from numpy.random import standard_normal
    >>>
    >>> X = standard_normal((100, 4))
    >>> np.allclose(trRV(X), (96.0, 96.0))
    True
    >>> V = np.identity(100)
    >>> np.allclose(trRV(X), (96.0, 96.0))
    True
    >>>
    >>> X[:,3] = X[:,1] + X[:,2]
    >>> np.allclose(trRV(X), (97.0, 97.0))
    True
    >>>
    >>> u = orth(X)
    >>> V = np.dot(u, u.T)
    >>> print np.allclose(trRV(X, V), 0)
    True
    """
    n, p = X.shape

    if V is None:
        V = np.identity(n)
        
    if X is None:
        if V is None:
            trRV = trRVRV = n
        else:
            trRV = _trace(V)
            trRVRV = _frobenius(V, V)
    else:
        u = orth(X)
        if V is None:
            trRV = trRVRV = n - u.shape[1]
        else:
            Vu = np.dot(V, u)
            utVu = np.dot(u.T, Vu)
            trRV = _trace(V) - _frobenius(u, Vu)
            trRVRV = _frobenius(V, V) - 2 * _frobenius(Vu, Vu) + _frobenius(utVu, utVu)

    return trRV, trRVRV

if __name__ == "__main__":

    from numpy.random import standard_normal

    X = standard_normal((100, 4))
    print trRV(X) # should be (96,96)

    V = np.identity(100)
    print trRV(X, V) # should be (96,96)

    X[:,3] = X[:,1] + X[:,2]
    print trRV(X, V) # should be (97,97)

    u = orth(X)
    V = np.dot(u, u.T)
    print trRV(X, V) # should be (0,0)
            



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
import numpy as np

from scipy.special import gammaln, hermitenorm
import scipy.stats
from scipy.misc import factorial

from .. import rft

from nose.tools import assert_raises

from numpy.testing import assert_almost_equal, dec

#def rho(x, dim, df=np.inf):
#    """
#    EC densities for T and Gaussian (df=inf) random fields. 
#    """
#
#    m = df
#    
#    if dim > 0:
#        x = np.asarray(x, np.float64)
#--jarrod: shouldn't Q be rft.Q??
#        q = Q(dim, dfd=df)(x)
#
#        if np.isfinite(m):
#            q *= np.power(1 + x**2/m, -(m-1)/2.)
#        else:
#            q *= np.exp(-x**2/2)
#
#        return q * np.power(2*np.pi, -(dim+1)/2.)
#    else:
#        if np.isfinite(m):
#            return scipy.stats.t.sf(x, df)
#        else:
#            return scipy.stats.norm.sf(x)

def test_Q():
    assert_raises(ValueError, rft.Q, -1)
    assert_raises(ValueError, rft.Q, 0)
    x = np.arange(-9, 10)
    for dim in range(1, 4):
        res = rft.Q(dim)
        assert_almost_equal(res(x), hermitenorm(dim - 1)(x))


def K(dim=4, dfn=7, dfd=np.inf):
    """
    Determine the polynomial K in:

        Worsley, K.J. (1994). 'Local maxima and the expected Euler
        characteristic of excursion sets of \chi^2, F and t fields.' Advances in
        Applied Probability, 26:13-42.

    If dfd=inf, return the limiting polynomial.
    """
    def lbinom(n, j):
        return gammaln(n+1) - gammaln(j+1) - gammaln(n-j+1)

    m = dfd
    n = dfn
    D = dim
    k = np.arange(D)
    coef = 0
    for j in range(int(np.floor((D-1)/2.)+1)):
        if np.isfinite(m):
            t = (gammaln((m+n-D)/2.+j) - 
                 gammaln(j+1) -
                 gammaln((m+n-D)/2.))
            t += lbinom(m-1, k-j) - k * np.log(m)
        else:
            _t = np.power(2., -j) / (factorial(k-j) * factorial(j))
            t = np.log(_t)
            t[np.isinf(_t)] = -np.inf
        t += lbinom(n-1, D-1-j-k) 
        coef += (-1)**(D-1) * factorial(D-1) * np.exp(t) * np.power(-1.*n, k) 
    return np.poly1d(coef[::-1])


def F(x, dim, dfd=np.inf, dfn=1):
    """
    EC densities for F and Chi^2 (dfd=inf) random fields. 
    """
    m = float(dfd)
    n = float(dfn)
    D = float(dim)
    if dim > 0:
        x = np.asarray(x, np.float64)
        k = K(dim=dim, dfd=dfd, dfn=dfn)(x)
        if np.isfinite(m):
            f = x*n/m
            t = -np.log(1 + f) * (m+n-2.) / 2.
            t += np.log(f) * (n-D) / 2.
            t += gammaln((m+n-D)/2.) - gammaln(m/2.)
        else:
            f = x*n
            t = np.log(f/2.) * (n-D) / 2. - f/2. 
        t -= np.log(2*np.pi) * D / 2. + np.log(2) * (D-2)/2. + gammaln(n/2.)
        k *= np.exp(t)
        return k
    else:
        if np.isfinite(m):
            return scipy.stats.f.sf(x, dfn, dfd)
        else:
            return scipy.stats.chi.sf(x, dfn)


def polyF(dim, dfd=np.inf, dfn=1):
    """
    Return the polynomial part of the EC density when evaluating the polynomial
    on the sqrt(F) scale (or sqrt(chi^2)=chi scale).

    The polynomial is such that, if dfd=inf, the F EC density in is just::

        polyF(dim,dfn=dfn)(sqrt(dfn*x)) * exp(-dfn*x/2) * (2\pi)^{-(dim+1)/2}
    """
    n = float(dfn)
    m = float(dfd)
    D = float(dim)
    p = K(dim=D, dfd=m, dfn=n)
    c = p.c
    # Take care of the powers of n (i.e. we want polynomial K evaluated
    # at */n).
    for i in range(p.order+1):
        c[i] /= np.power(n, p.order-i)
    # Now, turn it into a polynomial of x when evaluated at x**2
    C = np.zeros((2*c.shape[0]-1,))
    for i in range(c.shape[0]):
        C[2*i] = c[i]
    # Multiply by the factor x^(dfn-dim) in front (see Theorem 4.6 of
    # Worsley (1994), cited above.
    if dim > dfn: # divide by x^(dim-dfn)
        C = C[0:(C.shape[0] - (dim-dfn))]
    else: # multiply by x^(dim-dfn)
        C = np.hstack([C, np.zeros((dfn-dim,))])
    # Fix up constant in front
    if np.isfinite(m):
        C *= np.exp(gammaln((m+n-D)/2.) - gammaln(m/2.)) * np.power(m, -(n-D)/2.)
    else:
        C *= np.power(2, -(n-D)/2.)
    C /= np.power(2, (dim-2)/2.) * np.exp(gammaln(n/2.))
    C *= np.sqrt(2*np.pi)
    return np.poly1d(C)


def F_alternative(x, dim, dfd=np.inf, dfn=1):
    """
    Another way to compute F EC density as a product of a polynomial and a power
    of (1+x^2/m).
    """
    n = float(dfn)
    m = float(dfd)
    x = np.asarray(x, np.float64)
    p = polyF(dim=dim, dfd=dfd, dfn=dfn)
    v = p(np.sqrt(n*x))
    if np.isfinite(m):
        v *= np.power(1 + n*x/m, -(m+n-2.) / 2.)
    else:
        v *= np.exp(-n*x/2)
    v *= np.power(2*np.pi, -(dim+1)/2.)
    return v


def test_polynomial1():
    # Polynomial part of Gaussian densities are Hermite polynomials.
    for dim in range(1,10):
        q = rft.Gaussian().quasi(dim)
        h = hermitenorm(dim-1)
        yield assert_almost_equal, q.c, h.c


def test_polynomial2():
    # EC density of chi^2(1) is 2 * EC density of Gaussian so polynomial part is
    # a factor of 2 as well.
    for dim in range(1,10):
        q = rft.ChiSquared(dfn=1).quasi(dim)
        h = hermitenorm(dim-1)
        yield assert_almost_equal, q.c, 2*h.c


@dec.slow
def test_polynomial3():
    # EC density of F with infinite dfd is the same as chi^2 --
    # polynomials should be the same.
    for dim in range(10):
        for dfn in range(5,10):
            q1 = rft.FStat(dfn=dfn, dfd=np.inf).quasi(dim)
            q2 = rft.ChiSquared(dfn=dfn).quasi(dim)
            yield assert_almost_equal, q1.c, q2.c


@dec.slow
def test_chi1():
    # EC density of F with infinite dfd is the same as chi^2 -- EC should be the
    # same.
    x = np.linspace(0.1,10,100)
    for dim in range(10):
        for dfn in range(5,10):
            c = rft.ChiSquared(dfn=dfn)
            f = rft.FStat(dfn=dfn, dfd=np.inf)
            chi1 = c.density(dfn*x, dim)
            chi2 = f.density(x, dim)
            yield assert_almost_equal, chi1, chi2


def test_chi2():
    # Quasi-polynomial part of the chi^2 EC density should
    # be the limiting polyF.
    for dim in range(1,10):
        for dfn in range(5,10):
            c = rft.ChiSquared(dfn=dfn)
            p1 = c.quasi(dim=dim)
            p2 = polyF(dim=dim, dfn=dfn)
            yield assert_almost_equal, p1.c, p2.c


def test_chi3():
    # EC density of chi^2(1) is 2 * EC density of Gaussian squared so EC
    # densities factor of 2 as well.
    x = np.linspace(0.1,10,100)
    for dim in range(10):
        g = rft.Gaussian()
        c = rft.ChiSquared(dfn=1)
        ec1 = g.density(np.sqrt(x), dim)
        ec2 = c.density(x, dim)
        yield assert_almost_equal, 2*ec1, ec2


def test_T1():
    # O-dim EC density should be tail probality.
    x = np.linspace(0.1,10,100)
    for dfd in [40,50]:
        t = rft.TStat(dfd=dfd)
        yield assert_almost_equal, t(x), scipy.stats.t.sf(x, dfd)
    t = rft.TStat(dfd=np.inf)
    yield assert_almost_equal, t(x), scipy.stats.norm.sf(x)


def test_search():
    #  Test that the search region works.
    search = rft.IntrinsicVolumes([3,4,5])
    x = np.linspace(0.1,10,100)
    stat = rft.Gaussian(search=search)
    v1 = stat(x)
    v2 = ((5*x + 4*np.sqrt(2*np.pi)) *
          np.exp(-x**2/2.) / np.power(2*np.pi, 1.5) +
          3 * scipy.stats.norm.sf(x))
    assert_almost_equal(v1, v2)


@dec.slow
def test_search1():
    # Test that the search region works.
    # XXX - we are not testing anything
    search = rft.IntrinsicVolumes([3,4,5])
    x = np.linspace(0.1,10,100)
    stats = [rft.Gaussian()]
    for dfn in range(5,10):
        for dfd in [40,50,np.inf]:
            stats.append(rft.FStat(dfn=dfn, dfd=dfd))
            stats.append(rft.TStat(dfd=dfd))
        stats.append(rft.ChiSquared(dfn=dfn))
    for dim in range(7):
        for stat in stats:
            # XXX - v1 appears to be unused
            v1 = stat(x, search=search)
            v2 = 0
            for i in range(search.mu.shape[0]):
                v2 += stat.density(x, i) * search.mu[i]


@dec.slow
def test_search2():
    # Test that the search region works.
    search = rft.IntrinsicVolumes([3,4,5])
    x = np.linspace(0.1,10,100)
    stats = [rft.Gaussian(search=search)]
    ostats = [rft.Gaussian()]
    for dfn in range(5,10):
        for dfd in [40,50,np.inf]:
            stats.append(rft.FStat(dfn=dfn, dfd=dfd, search=search))
            ostats.append(rft.FStat(dfn=dfn, dfd=dfd))
            stats.append(rft.TStat(dfd=dfd, search=search))
            ostats.append(rft.TStat(dfd=dfd))
        stats.append(rft.ChiSquared(dfn=dfn, search=search))
        ostats.append(rft.ChiSquared(dfn=dfn))
    for i in range(len(stats)):
        stat = stats[i]
        ostat = ostats[i]
        v1 = stat(x)
        v2 = 0
        for j in range(search.mu.shape[0]):
            v2 += ostat.density(x, j) * search.mu[j]
        assert_almost_equal(v1, v2)


def test_search3():
    # In the Gaussian case, test that search and product give same results.
    search = rft.IntrinsicVolumes([3,4,5,7])
    g1 = rft.Gaussian(search=search)
    g2 = rft.Gaussian(product=search)
    x = np.linspace(0.1,10,100)
    y1 = g1(x)
    y2 = g2(x)
    assert_almost_equal(y1, y2)


def test_search4():
    # Test that the search/product work well together
    search = rft.IntrinsicVolumes([3,4,5])
    product = rft.IntrinsicVolumes([1,2])
    x = np.linspace(0.1,10,100)
    g1 = rft.Gaussian()
    g2 = rft.Gaussian(product=product)
    y = g2(x, search=search)
    z = g1(x, search=search*product)
    assert_almost_equal(y, z)


def test_search5():
    # Test that the search/product work well together
    search = rft.IntrinsicVolumes([3,4,5])
    product = rft.IntrinsicVolumes([1,2])
    prodsearch = product * search
    x = np.linspace(0,5,101)
    g1 = rft.Gaussian()
    g2 = rft.Gaussian(product=product)
    z = 0
    for i in range(prodsearch.mu.shape[0]):
        z += g1.density(x, i) * prodsearch.mu[i]
    y = g2(x, search=search)
    assert_almost_equal(y, z)


@dec.slow
def test_T2():
    # T**2 is an F with dfn=1
    x = np.linspace(0,5,101)
    for dfd in [40,50,np.inf]:
        t = rft.TStat(dfd=dfd)
        f = rft.FStat(dfd=dfd, dfn=1)
        for dim in range(7):
            y = 2*t.density(x, dim)
            z = f.density(x**2, dim)
            yield assert_almost_equal, y, z


@dec.slow
def test_hotelling1():
    # Asymptotically, Hotelling is the same as F which is the same as chi^2.
    x = np.linspace(0.1,10,100)
    for dim in range(7):
        for dfn in range(5,10):
            h = rft.Hotelling(k=dfn).density(x*dfn, dim)
            f = rft.FStat(dfn=dfn).density(x, dim)
            yield assert_almost_equal, h, f


@dec.slow
def test_hotelling4():
    # Hotelling T^2 should just be like taking product with sphere.
    x = np.linspace(0.1,10,100)
    for dim in range(7):
        search = rft.IntrinsicVolumes([0]*(dim) + [1])
        for k in range(5, 10):
            p = rft.spherical_search(k)
            for dfd in [np.inf,40,50]:
                f = rft.FStat(dfd=dfd, dfn=1)(x, search=p*search)
                t = 2*rft.TStat(dfd=dfd)(np.sqrt(x), search=p*search)
                h2 = 2*rft.Hotelling(k=k, dfd=dfd).density(x, dim)
                h = 2*rft.Hotelling(k=k, dfd=dfd)(x, search=search)

                yield assert_almost_equal, h, t
                yield assert_almost_equal, h, f
                yield assert_almost_equal, h, h2
    search = rft.IntrinsicVolumes([3,4,5])
    for k in range(5, 10):
        p = rft.spherical_search(k)
        for dfd in [np.inf,40,50]:
            f = rft.FStat(dfd=dfd, dfn=1)(x, search=p*search)
            h = 2*rft.Hotelling(k=k, dfd=dfd)(x, search=search)

            h2 = 0
            for i in range(search.mu.shape[0]):
                h2 += 2*rft.Hotelling(k=k, dfd=dfd).density(x, i) * search.mu[i]
            yield assert_almost_equal, h, f
            yield assert_almost_equal, h, h2


def test_hotelling2():
    # Marginally, Hotelling's T^2(k) with m degrees of freedom
    # in the denominator satisfies
    # (m-k+1)/(mk) T^2 \sim  F_{k,m-k+1}.
    x = np.linspace(0.1,10,100)
    for dfn in range(6, 10):
        h = rft.Hotelling(k=dfn)(x)
        chi = rft.ChiSquared(dfn=dfn)(x)
        assert_almost_equal(h, chi)
        chi2 = scipy.stats.chi2.sf(x, dfn)
        yield assert_almost_equal, h, chi2
        # XXX - p appears to be unused
        p = rft.spherical_search(dfn)
        for dfd in [40,50]:
            fac = (dfd-dfn+1.)/(dfd*dfn)
            h = rft.Hotelling(dfd=dfd,k=dfn)(x)
            f = scipy.stats.f.sf(x*fac, dfn, dfd-dfn+1)
            f2 = rft.FStat(dfd=dfd-dfn+1,dfn=dfn)(x*fac)
            yield assert_almost_equal, f2, f
            yield assert_almost_equal, h, f


@dec.slow
def test_roy1():
    # EC densities of Roy with dfn=1 should be twice EC densities of Hotelling
    # T^2's.
    x = np.linspace(0.1,10,100)
    for dfd in [40,50,np.inf]:
        for k in [1,4,6]:
            for dim in range(7):
                h = 2*rft.Hotelling(dfd=dfd,k=k).density(x, dim)
                r = rft.Roy(dfd=dfd,k=k,dfn=1).density(x, dim)
                yield assert_almost_equal, h, r


@dec.slow
def test_onesidedF():
    # EC densities of one sided F should be a difference of
    # F EC densities
    x = np.linspace(0.1,10,100)
    for dfd in [40,50,np.inf]:
        for dfn in range(2,10):
            for dim in range(7):
                f1 = rft.FStat(dfd=dfd,dfn=dfn).density(x, dim)
                f2 = rft.FStat(dfd=dfd,dfn=dfn-1).density(x, dim)
                onesided = rft.OneSidedF(dfd=dfd,dfn=dfn).density(x, dim)
                yield assert_almost_equal, onesided, 0.5*(f1-f2)


@dec.slow
def test_multivariate_forms():
    # MVform with one sphere is sqrt(chi^2), two spheres is sqrt(Roy) with
    # infinite degrees of freedom.
    x = np.linspace(0.1,10,100)
    for k1 in range(5,10):
        m = rft.MultilinearForm(k1)
        c = rft.ChiSquared(k1)
        for dim in range(7):
            mx = m.density(x, dim)
            cx = c.density(x**2, dim)
            yield assert_almost_equal, mx, cx
        for k2 in range(5,10):
            m = rft.MultilinearForm(k1,k2)
            r = rft.Roy(k=k1, dfn=k2, dfd=np.inf)
            for dim in range(7):
                mx = 2*m.density(x, dim)
                rx = r.density(x**2/k2, dim)
                yield assert_almost_equal, mx, rx


def test_scale():
    # Smoke test?
    a = rft.IntrinsicVolumes([2,3,4])
    b = rft.scale_space(a, [3,4], kappa=0.5)


def test_F1():
    x = np.linspace(0.1,10,100)
    for dim in range(1,10):
        for dfn in range(5,10):
            for dfd in [40,50,np.inf]:
                f1 = F(x, dim, dfn=dfn, dfd=dfd)
                f2 = F_alternative(x, dim, dfn=dfn, dfd=dfd)
                yield assert_almost_equal, f1, f2


@dec.slow
def test_F2():
    x = np.linspace(0.1,10,100)
    for dim in range(3,7):
        for dfn in range(5,10):
            for dfd in [40,50,np.inf]:
                f1 = rft.FStat(dfn=dfn, dfd=dfd).density(x, dim) 
                f2 = F_alternative(x, dim, dfn=dfn, dfd=dfd)
                yield assert_almost_equal, f1, f2


@dec.slow
def test_F3():
    x = np.linspace(0.1,10,100)
    for dim in range(3,7):
        for dfn in range(5,10):
            for dfd in [40,50,np.inf]:
                f1 = rft.FStat(dfn=dfn, dfd=dfd).density(x, dim) 
                f2 = F(x, dim, dfn=dfn, dfd=dfd)
                yield assert_almost_equal, f1, f2



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" The io.files module provides basic functions for working with file-based
images in nipy.

* load : load an image from a file
* save : save an image to a file

Examples
--------
See documentation for load and save functions for worked examples.
"""

import os

import numpy as np

import nibabel as nib
from nibabel.spatialimages import HeaderDataError

from ..core.image.image import is_image

from .nifti_ref import (nipy2nifti, nifti2nipy)


def load(filename):
    """Load an image from the given filename.

    Parameters
    ----------
    filename : string
        Should resolve to a complete filename path.

    Returns
    -------
    image : An `Image` object
        If successful, a new `Image` object is returned.

    See Also
    --------
    save_image : function for saving images
    Image : image object

    Examples
    --------
    >>> from nipy.io.api import load_image
    >>> from nipy.testing import anatfile
    >>> img = load_image(anatfile)
    >>> img.shape
    (33, 41, 25)
    """
    img = nib.load(filename)
    ni_img = nib.Nifti1Image(img._data, img.get_affine(), img.get_header())
    return nifti2nipy(ni_img)


def save(img, filename, dtype_from='data'):
    """Write the image to a file.

    Parameters
    ----------
    img : An `Image` object
    filename : string
        Should be a valid filename.
    dtype_from : {'data', 'header'} or dtype specifier, optional
        Method of setting dtype to save data to disk. Value of 'data' (default),
        means use data dtype to save.  'header' means use data dtype specified
        in header, if available, otherwise use data dtype.  Can also be any
        valid specifier for a numpy dtype, e.g. 'i4', ``np.float32``.  Not every
        format supports every dtype, so some values of this parameter or data
        dtypes will raise errors.

    Returns
    -------
    image : An `Image` object
        Possibly modified by saving.

    See Also
    --------
    load_image : function for loading images
    Image : image object

    Examples
    --------
    Make a temporary directory to store files

    >>> import os
    >>> from tempfile import mkdtemp
    >>> tmpdir = mkdtemp()

    Make some some files and save them

    >>> import numpy as np
    >>> from nipy.core.api import Image, AffineTransform
    >>> from nipy.io.api import save_image
    >>> data = np.zeros((91,109,91), dtype=np.uint8)
    >>> cmap = AffineTransform('kji', 'zxy', np.eye(4))
    >>> img = Image(data, cmap)
    >>> fname1 = os.path.join(tmpdir, 'img1.nii.gz')
    >>> saved_img1 = save_image(img, fname1)
    >>> saved_img1.shape
    (91, 109, 91)
    >>> fname2 = os.path.join(tmpdir, 'img2.img.gz')
    >>> saved_img2 = save_image(img, fname2)
    >>> saved_img2.shape
    (91, 109, 91)
    >>> fname = 'test.mnc'
    >>> saved_image3 = save_image(img, fname)
    Traceback (most recent call last):
       ...
    ValueError: Sorry, we cannot yet save as format "minc"

    Finally, we clear up our temporary files:

    >>> import shutil
    >>> shutil.rmtree(tmpdir)

    Notes
    -----
    Filetype is determined by the file extension in 'filename'.  Currently the
    following filetypes are supported:

    * Nifti single file : ['.nii', '.nii.gz']
    * Nifti file pair : ['.hdr', '.hdr.gz']
    * SPM Analyze : ['.img', '.img.gz']
    """
    # Try and get nifti
    dt_from_is_str = isinstance(dtype_from, basestring)
    if dt_from_is_str and dtype_from == 'header':
        # All done
        io_dtype = None
    elif dt_from_is_str and dtype_from == 'data':
        io_dtype = img.get_data().dtype
    else:
        io_dtype = np.dtype(dtype_from)
    # make new image
    ni_img = nipy2nifti(img, data_dtype = io_dtype)
    ftype = _type_from_filename(filename)
    if ftype.startswith('nifti1'):
        ni_img.to_filename(filename)
    elif ftype == 'analyze':
        try:
            ana_img = nib.Spm2AnalyzeImage.from_image(ni_img)
        except HeaderDataError:
            raise HeaderDataError('SPM analyze does not support datatype %s' %
                                  ni_img.get_header().get_data_dtype())
        ana_img.to_filename(filename)
    else:
        raise ValueError('Sorry, we cannot yet save as format "%s"' % ftype)
    return img


def _type_from_filename(filename):
    ''' Return image type determined from filename

    Filetype is determined by the file extension in 'filename'.
    Currently the following filetypes are supported:

    * Nifti single file : ['.nii', '.nii.gz']
    * Nifti file pair : ['.hdr', '.hdr.gz']
    * Analyze file pair : ['.img', '.img.gz']

    >>> _type_from_filename('test.nii')
    'nifti1single'
    >>> _type_from_filename('test')
    'nifti1single'
    >>> _type_from_filename('test.hdr')
    'nifti1pair'
    >>> _type_from_filename('test.hdr.gz')
    'nifti1pair'
    >>> _type_from_filename('test.img.gz')
    'analyze'
    >>> _type_from_filename('test.mnc')
    'minc'
    '''
    if filename.endswith('.gz'):
        filename = filename[:-3]
    elif filename.endswith('.bz2'):
        filename = filename[:-4]
    _, ext = os.path.splitext(filename)
    if ext in ('', '.nii'):
        return 'nifti1single'
    if ext == '.hdr':
        return 'nifti1pair'
    if ext == '.img':
        return 'analyze'
    if ext == '.mnc':
        return 'minc'
    raise ValueError('Strange file extension "%s"' % ext)


def as_image(image_input):
    ''' Load image from filename or pass through image instance

    Parameters
    ----------
    image_input : str or Image instance
       image or string filename of image.  If a string, load image and
       return.  If an image, pass through without modification

    Returns
    -------
    img : Image or Image-like instance
       Input object if `image_input` seemed to be an image, loaded Image
       object if `image_input` was a string.

    Raises
    ------
    TypeError : if neither string nor image-like passed

    Examples
    --------
    >>> from nipy.testing import anatfile
    >>> from nipy.io.api import load_image
    >>> img = as_image(anatfile)
    >>> img2 = as_image(img)
    >>> img2 is img
    True
    '''
    if is_image(image_input):
        return image_input
    if isinstance(image_input, basestring):
        return load(image_input)
    raise TypeError('Expecting an image-like object or filename string')



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Perform a global search and replace on the current directory *recursively*.

This a small python wrapper around the `perl -p -i -e` functionality.
I **strongly recommend** running `perlpie` on files under source
control.  In this way it's easy to track your changes and if you
discover your regular expression was wrong you can easily revert.  I
also recommend using `grin` to test your regular expressions before
running `perlpie`.

Parameters
----------
regex : regular expression
    Regular expression matching the string you want to replace
newstring : string
    The string you would like to replace the oldstring with.  Note
    this is not a regular expression but the exact string. One
    exception to this rule is the at symbol `@`.  This has special
    meaning in perl, so you need an escape character for this.  See
    Examples below.

Requires
--------
perl : The underlying language we're using to perform the search and replace.

`grin <http://pypi.python.org/pypi/grin/>`_ : Grin is a tool written
by Robert Kern to wrap `grep` and `find` with python and easier
command line options.

Examples
--------
Replace all occurences of foo with bar::

    perlpie foo bar

Replace numpy.testing with nipy's testing framework::

    perlpie 'from\s+numpy\.testing.*' 'from nipy.testing import *'

Replace all @slow decorators in my code with @dec.super_slow.  Here we
have to escape the @ symbol which has special meaning in perl::
    
    perlpie '\@slow' '\@dec.super_slow'

Remove all occurences of importing make_doctest_suite::

    perlpie 'from\snipy\.utils\.testutils.*make_doctest_suite'

"""

# notes on perl-dash-pie
# perl -p -i -e 's/oldstring/newstring/g' *
# find . -name '*.html' -print0 | xargs -0 perl -pi -e 's/oldstring/newstring/g'

from optparse import OptionParser
import subprocess

usage_doc = "usage: %prog [options] regex newstring"

def check_deps():
    try:
        import grin
    except ImportError:
        print 'perlpie requires grin to operate.'
        print 'You can find grin in the python package index:'
        print '  http://pypi.python.org/pypi/grin/'
        return False
    # assume they have perl for now
    return True


def perl_dash_pie(oldstr, newstr, dry_run=None):
    """Use perl to replace the oldstr with the newstr.

    Examples
    --------

    # To replace all occurences of 'import numpy as N' with 'import numpy as np'
    from nipy.utils import perlpie
    perlpie.perl_dash_pie('import\s+numpy\s+as\s+N', 'import numpy as np')
    grind | xargs perl -pi -e 's/import\s+numpy\s+as\s+N/import numpy as np/g'
    
    """

    if dry_run:
        cmd = "grind | xargs perl -p -e 's/%s/%s/g'" % (oldstr, newstr)
    else:
        cmd = "grind | xargs perl -pi -e 's/%s/%s/g'" % (oldstr, newstr)
    print cmd

    try:
        subprocess.check_call(cmd, shell=True)
    except subprocess.CalledProcessError, err:
        msg = """
        Error while executing perl_dash_pie command:
        %s
        Error:
        %s
        """ % (cmd, str(err))
        raise Exception(msg)

def print_extended_help(option, opt_str, value, parser, *args, **kwargs):
    print __doc__

def main():
    description = __doc__.splitlines()[0]
    usage = usage_doc
    parser = OptionParser(usage=usage, description=description)
    parser.add_option('-e', '--extended-help', action='callback', 
                      callback=print_extended_help,
                      help='print extended help including examples')
    parser.add_option('-n', '--dry-run', action="store_true", dest="dry_run",
                      help='send results to stdout without modifying files')
    (options, args) = parser.parse_args()

    if not args:
        parser.print_help()
        return

    if check_deps():
        oldstr = args[0]
        newstr = args[1]
        perl_dash_pie(oldstr, newstr, options.dry_run)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
""" Random field theory routines

The theoretical results for  the EC densities appearing in this module
were partially supported by NSF grant DMS-0405970.

Taylor, J.E. & Worsley, K.J. (2007).  "Detecting sparse cone alternatives
   for Gaussian random fields, with an application to fMRI".
   Annals of Statistics, submitted.

Taylor, J.E. & Worsley, K.J. (2007). "Random fields of multivariate
   test statistics, with applications to shape analysis."
   Annals of Statistics, accepted.
"""

import numpy as np
from numpy.linalg import pinv    

from scipy import stats
from scipy.misc import factorial
from scipy.special import gamma, gammaln, beta, hermitenorm

def binomial(n, k):
    """ Binomial coefficient

               n!
    c =    ---------
           (n-k)! k!

    Parameters
    ----------
    n : float
       n of (n, k)
    k : float
       k of (n, k)

    Returns
    -------
    c : float

    Examples
    --------
    First 3 values of 4 th row of Pascal triangle

    >>> [binomial(4, k) for k in range(3)]
    [1.0, 4.0, 6.0]
    """
    if n <= k or n == 0:
        return 0.
    elif k == 0:
        return 1.
    return 1./(beta(n-k+1, k+1)*(n+1))


def Q(dim, dfd=np.inf):
    """ Q polynomial

    If `dfd` == inf (the default), then Q(dim) is the (dim-1)-st Hermite
    polynomial:

    .. math::

        H_j(x) = (-1)^j * e^{x^2/2} * (d^j/dx^j e^{-x^2/2})

    If `dfd` != inf, then it is the polynomial Q defined in [Worsley1994]_

    Parameters
    ----------
    dim : int
        dimension of polynomial
    dfd : scalar

    Returns
    -------
    q_poly : np.poly1d instance

    References
    ----------
    .. [Worsley1994] Worsley, K.J. (1994). 'Local maxima and the expected Euler
       characteristic of excursion sets of \chi^2, F and t fields.' Advances in
       Applied Probability, 26:13-42.
    """
    m = dfd
    j = dim
    if j <= 0:
        raise ValueError('Q defined only for dim > 0')
    poly = hermitenorm(j-1)
    poly = np.poly1d(np.around(poly.c))
    if np.isfinite(m):
        for l in range((j-1)//2+1):
            f = np.exp(gammaln((m+1)/2.) - gammaln((m+2-j+2*l)/2.)
                                - 0.5*(j-1-2*l)*(np.log(m/2.)))
            poly.c[2*l] *= f
    return np.poly1d(poly.c)


class ECquasi(np.poly1d):
    """ Polynomials with premultiplier

    A subclass of poly1d consisting of polynomials with a premultiplier of the
    form:

    (1 + x^2/m)^-exponent

    where m is a non-negative float (possibly infinity, in which case the
    function is a polynomial) and exponent is a non-negative multiple of 1/2.

    These arise often in the EC densities.

    Examples
    --------
    >>> import numpy
    >>> from nipy.algorithms.statistics.rft import ECquasi
    >>> x = numpy.linspace(0,1,101)

    >>> a = ECquasi([3,4,5])
    >>> a
    ECquasi(array([3, 4, 5]), m=inf, exponent=0.000000)
    >>> a(3) == 3*3**2 + 4*3 + 5
    True

    >>> b = ECquasi(a.coeffs, m=30, exponent=4)
    >>> numpy.allclose(b(x), a(x) * numpy.power(1+x**2/30, -4))
    True
    """
    def __init__(self, c_or_r, r=0, exponent=None, m=None):
        np.poly1d.__init__(self, c_or_r, r=r, variable='x')
        if exponent is None and not hasattr(self, 'exponent'):
            self.exponent = 0
        elif not hasattr(self, 'exponent'):
            self.exponent = exponent
        if m is None and not hasattr(self, 'm'):
            self.m = np.inf
        elif not hasattr(self, 'm'):
            self.m = m
        if not np.isfinite(self.m):
            self.exponent = 0.

    def denom_poly(self):
        """ Base of the premultiplier: (1+x^2/m).

        Examples
        --------
        >>> import numpy
        >>> b = ECquasi([3,4,20], m=30, exponent=4)
        >>> d = b.denom_poly()
        >>> d
        poly1d([ 0.03333333,  0.        ,  1.        ])
        >>> numpy.allclose(d.c, [1./b.m,0,1])
        True
        """
        return np.poly1d([1./self.m, 0, 1])

    def change_exponent(self, _pow):
        """ Change exponent

        Multiply top and bottom by an integer multiple of the
        self.denom_poly.

        Examples
        --------
        >>> import numpy
        >>> b = ECquasi([3,4,20], m=30, exponent=4)
        >>> x = numpy.linspace(0,1,101)
        >>> c = b.change_exponent(3)
        >>> c
        ECquasi(array([  1.11111111e-04,   1.48148148e-04,   1.07407407e-02,
                 1.33333333e-02,   3.66666667e-01,   4.00000000e-01,
                 5.00000000e+00,   4.00000000e+00,   2.00000000e+01]), m=30.000000, exponent=7.000000)
        >>> numpy.allclose(c(x), b(x))
        True
        """
        if np.isfinite(self.m):
            _denom_poly = self.denom_poly()
            if int(_pow) != _pow or _pow < 0:
                raise ValueError, 'expecting a non-negative integer'
            p = _denom_poly**int(_pow)
            exponent = self.exponent + _pow
            coeffs = np.polymul(self, p).coeffs
            return ECquasi(coeffs, exponent=exponent, m=self.m)
        else:
            return ECquasi(self.coeffs, exponent=self.exponent, m=self.m)

    def __setattr__(self, key, val):
        if key == 'exponent':
            if 2*float(val) % 1 == 0: 
                self.__dict__[key] = float(val)
            else: raise ValueError, 'expecting multiple of a half, got %f' % val
        elif key == 'm':
            if float(val) > 0 or val == np.inf:
                self.__dict__[key] = val
            else: raise ValueError, 'expecting positive float or inf'
        else: np.poly1d.__setattr__(self, key, val)

    def compatible(self, other):
        """ Check compatibility of degrees of freedom

        Check whether the degrees of freedom of two instances are equal
        so that they can be multiplied together.

        Examples
        --------
        >>> import numpy
        >>> b = ECquasi([3,4,20], m=30, exponent=4)
        >>> x = numpy.linspace(0,1,101)
        >>> c = b.change_exponent(3)
        >>> b.compatible(c)
        True
        >>> d = ECquasi([3,4,20])
        >>> b.compatible(d)
        False
        >>>
        """
        if self.m != other.m:
            #raise ValueError, 'quasi polynomials are not compatible, m disagrees'
            return False
        return True

    def __add__(self, other):
        """ Add two compatible ECquasi instances together.

        Examples
        --------
        >>> b = ECquasi([3,4,20], m=30, exponent=4)
        >>> c = ECquasi([1], m=30, exponent=4)
        >>> b+c #doctest: +IGNORE_DTYPE
        ECquasi(array([ 3,  4, 21]), m=30.000000, exponent=4.000000)

        >>> d = ECquasi([1], m=30, exponent=3)
        >>> b+d
        ECquasi(array([  3.03333333,   4.        ,  21.        ]), m=30.000000, exponent=4.000000)
        """
        if self.compatible(other):
            if np.isfinite(self.m):
                M = max(self.exponent, other.exponent)
                q1 = self.change_exponent(M-self.exponent)
                q2 = other.change_exponent(M-other.exponent)
                p = np.poly1d.__add__(q1, q2)
                return ECquasi(p.coeffs,
                               exponent=M,
                               m=self.m)
            else:
                p = np.poly1d.__add__(self, other)
                return ECquasi(p.coeffs,
                               exponent=0,
                               m=self.m)

    def __mul__(self, other):
        """ Multiply two compatible ECquasi instances together.

        Examples
        --------
        >>> b=ECquasi([3,4,20], m=30, exponent=4)
        >>> c=ECquasi([1,2], m=30, exponent=4.5)
        >>> b*c
        ECquasi(array([ 3, 10, 28, 40]), m=30.000000, exponent=8.500000)
        """
        if np.isscalar(other):
            return ECquasi(self.coeffs * other,
                           m=self.m,
                           exponent=self.exponent)
        elif self.compatible(other):
            p = np.poly1d.__mul__(self, other)
            return ECquasi(p.coeffs,
                           exponent=self.exponent+other.exponent,
                           m=self.m)

    def __call__(self, val):
        """Evaluate the ECquasi instance.

        Examples
        --------
        >>> import numpy
        >>> x = numpy.linspace(0,1,101)
        >>> a = ECquasi([3,4,5])
        >>> a
        ECquasi(array([3, 4, 5]), m=inf, exponent=0.000000)
        >>> a(3) == 3*3**2 + 4*3 + 5
        True
        >>> b = ECquasi(a.coeffs, m=30, exponent=4)
        >>> numpy.allclose(b(x), a(x) * numpy.power(1+x**2/30, -4))
        True
        """
        n = np.poly1d.__call__(self, val)
        _p = self.denom_poly()(val)
        return n / np.power(_p, self.exponent)

    def __div__(self, other):
        raise NotImplementedError

    def __eq__(self, other):
        return (np.poly1d.__eq__(self, other) and
                self.m == other.m and
                self.exponent == other.exponent)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __pow__(self, _pow):
        """ Power of a ECquasi instance.

        Examples
        --------
        >>> b = ECquasi([3,4,5],m=10, exponent=3)
        >>> b**2
        ECquasi(array([ 9, 24, 46, 40, 25]), m=10.000000, exponent=6.000000)
        """
        p = np.poly1d.__pow__(self, int(_pow))
        q = ECquasi(p, m=self.m, exponent=_pow*self.exponent)
        return q

    def __sub__(self, other):
        """ Subtract `other` from `self`

        Parameters
        ----------
        other : ECquasi instance

        Returns
        -------
        subbed : ECquasi

        Examples
        --------
        >>> b = ECquasi([3,4,20], m=30, exponent=4)
        >>> c = ECquasi([1,2], m=30, exponent=4)
        >>> print b-c #doctest: +IGNORE_DTYPE
        ECquasi(array([ 3,  3, 18]), m=30.000000, exponent=4.000000)
        """
        return self + (other * -1)

    def __repr__(self):
        if not np.isfinite(self.m):
            m = repr(self.m)
        else:
            m = '%f' % self.m
        return "ECquasi(%s, m=%s, exponent=%f)" % (
            repr(self.coeffs), m, self.exponent)

    __str__ = __repr__
    __rsub__ = __sub__
    __rmul__ = __mul__
    __rdiv__ = __div__

    def deriv(self, m=1):
        """ Evaluate derivative of ECquasi

        Parameters
        ----------
        m : int, optional

        Examples
        --------
        >>> a = ECquasi([3,4,5])
        >>> a.deriv(m=2) #doctest: +IGNORE_DTYPE
        ECquasi(array([6]), m=inf, exponent=0.000000)

        >>> b = ECquasi([3,4,5], m=10, exponent=3)
        >>> b.deriv()
        ECquasi(array([-1.2, -2. ,  3. ,  4. ]), m=10.000000, exponent=4.000000)
        """
        if m == 1:
            if np.isfinite(self.m):
                q1 = ECquasi(np.poly1d.deriv(self, m=1),
                             m=self.m,
                             exponent=self.exponent)
                q2 = ECquasi(np.poly1d.__mul__(self, self.denom_poly().deriv(m=1)),
                             m = self.m,
                             exponent=self.exponent+1)
                return q1 - self.exponent * q2
            else:
                return ECquasi(np.poly1d.deriv(self, m=1),
                               m=np.inf,
                               exponent=0)
        else:
            d = self.deriv(m=1)
            return d.deriv(m=m-1)


class fnsum(object):
    def __init__(self, *items):
        self.items = list(items)

    def __call__(self, x):
        v = 0
        for q in self.items:
            v += q(x)
        return v


class IntrinsicVolumes(object):
    """ Compute intrinsic volumes of products of sets

    A simple class that exists only to compute the intrinsic volumes of
    products of sets (that themselves have intrinsic volumes, of course).
    """
    def __init__(self, mu=[1]):
        if isinstance(mu, IntrinsicVolumes):
            mu = mu.mu
        self.mu = np.asarray(mu, np.float64)
        self.order = self.mu.shape[0]-1

    def __str__(self):
        return str(self.mu)

    def __mul__(self, other):
        if not isinstance(other, IntrinsicVolumes):
            raise ValueError('expecting an IntrinsicVolumes instance')
        order = self.order + other.order + 1
        mu = np.zeros(order)

        for i in range(order):
            for j in range(i+1):
                try:
                    mu[i] += self.mu[j] * other.mu[i-j]
                except:
                    pass
        return self.__class__(mu)


class ECcone(IntrinsicVolumes):
    """ EC approximation to supremum distribution of var==1 Gaussian process

    A class that takes the intrinsic volumes of a set and gives the EC
    approximation to the supremum distribution of a unit variance Gaussian
    process with these intrinsic volumes. This is the basic building block of
    all of the EC densities.

    If product is not None, then this product (an instance of IntrinsicVolumes)
    will effectively be prepended to the search region in any call, but it will
    also affect the (quasi-)polynomial part of the EC density. For instance,
    Hotelling's T^2 random field has a sphere as product, as does Roy's maximum
    root.
    """
    def __init__(self, mu=[1], dfd=np.inf, search=[1], product=[1]):
        self.dfd = dfd
        IntrinsicVolumes.__init__(self, mu=mu)
        self.product = IntrinsicVolumes(product)
        self.search = IntrinsicVolumes(search)

    def __call__(self, x, search=None):
        """ Get expected EC for a search region

        Default is self.search which itself defaults to [1] giving the
        survival function.
        """
        x = np.asarray(x, np.float64)
        if search is None:
            search = self.search
        else:
            search = IntrinsicVolumes(search)

        search *= self.product

        if np.isfinite(self.dfd):
            q_even = ECquasi([0], m=self.dfd, exponent=0)
            q_odd = ECquasi([0], m=self.dfd, exponent=0.5)
        else:
            q_even = np.poly1d([0])
            q_odd = np.poly1d([0])

        for k in range(search.mu.shape[0]):
            q = self.quasi(k)
            c = float(search.mu[k]) * np.power(2*np.pi, -(k+1)/2.)
            if np.isfinite(self.dfd):
                q_even += q[0] * c
                q_odd += q[1] * c
            else:
                q_even += q * c

        _rho = q_even(x) + q_odd(x)

        if np.isfinite(self.dfd):
            _rho *= np.power(1 + x**2/self.dfd, -(self.dfd-1)/2.)
        else:
            _rho *= np.exp(-x**2/2.)

        if search.mu[0] * self.mu[0] != 0.:
            # tail probability is not "quasi-polynomial"
            if not np.isfinite(self.dfd):
                P = stats.norm.sf
            else:
                P = lambda x: stats.t.sf(x, self.dfd)
            _rho += P(x) * search.mu[0] * self.mu[0]
        return _rho

    def pvalue(self, x, search=None):
        return self(x, search=search)

    def integ(self, m=None, k=None):
        raise NotImplementedError # this could be done with stats.t,
                                  # at least m=1
    
    def density(self, x, dim):
        """ The EC density in dimension `dim`.
        """
        return self(x, search=[0]*dim+[1])

    def _quasi_polynomials(self, dim):
        """ list of quasi-polynomials for EC density calculation.
        """
        c = self.mu / np.power(2*np.pi, np.arange(self.order+1.)/2.)

        quasi_polynomials = []

        for k in range(c.shape[0]):
            if k+dim > 0:
                _q = ECquasi(Q(k+dim, dfd=self.dfd),
                             m=self.dfd,
                             exponent=k/2.)
                _q *= float(c[k])
                quasi_polynomials.append(_q)
        return quasi_polynomials

    def quasi(self, dim):
        """ (Quasi-)polynomial parts of EC density in dimension `dim`

        - ignoring a factor of (2\pi)^{-(dim+1)/2} in front.
        """
        q_even = ECquasi([0], m=self.dfd, exponent=0)
        q_odd = ECquasi([0], m=self.dfd, exponent=0.5)

        quasi_polynomials = self._quasi_polynomials(dim)
        for k in range(len(quasi_polynomials)):
            _q = quasi_polynomials[k]
            if _q.exponent % 1 == 0:
                q_even += _q
            else:
                q_odd += _q

        if not np.isfinite(self.dfd):
            q_even += q_odd
            return np.poly1d(q_even.coeffs)

        else:
            return (q_even, q_odd)

Gaussian = ECcone


def mu_sphere(n, j, r=1):
    """ `j`th curvature for `n` dimensional sphere radius `r`

    Return mu_j(S_r(R^n)), the j-th Lipschitz Killing
    curvature of the sphere of radius r in R^n.

    From Chapter 6 of

    Adler & Taylor, 'Random Fields and Geometry'. 2006.
    """
    if j < n:
        if n-1 == j:
            return 2 * np.power(np.pi, n/2.) * np.power(r, n-1) / gamma(n/2.) 

        if (n-1-j)%2 == 0:

            return 2 * binomial(n-1, j) * mu_sphere(n,n-1) * np.power(r, j) / mu_sphere(n-j,n-j-1)
        else:
            return 0
    else:
        return 0


def mu_ball(n, j, r=1):
    """ `j`th curvature of `n`-dimensional ball radius `r`

    Return mu_j(B_n(r)), the j-th Lipschitz Killing curvature of the
    ball of radius r in R^n.
    """
    if j <= n:
        if n == j:
            return np.power(np.pi, n/2.) * np.power(r, n) / gamma(n/2. + 1.)
        else:
            return binomial(n, j) * np.power(r, j) * mu_ball(n,n) / mu_ball(n-j,n-j)
    else:
        return 0


def spherical_search(n, r=1):
    """ A spherical search region of radius r.
    """
    return IntrinsicVolumes([mu_sphere(n,j,r=r) for j in range(n)])


def ball_search(n, r=1):
    """ A ball-shaped search region of radius r.
    """
    return IntrinsicVolumes([mu_ball(n,j,r=r) for j in range(n+1)])


def volume2ball(vol, d=3):
    """ Approximate volume with ball

    Approximate intrinsic volumes of a set with a given volume by those of a
    ball with a given dimension and equal volume.
    """
    if d > 0:
        r = np.power(vol * 1. / mu_ball(d, d), 1./d)
        return ball_search(d, r=r)
    else:
        return IntrinsicVolumes([1])


class ChiSquared(ECcone):
    """  EC densities for a Chi-Squared(n) random field.
    """
    def __init__(self, dfn, dfd=np.inf, search=[1]):
        self.dfn = dfn
        ECcone.__init__(self, mu=spherical_search(self.dfn), search=search, dfd=dfd)

    def __call__(self, x, search=None):
        return ECcone.__call__(self, np.sqrt(x), search=search)


class TStat(ECcone):
    """  EC densities for a t random field.
    """
    def __init__(self, dfd=np.inf, search=[1]):
        ECcone.__init__(self, mu=[1], dfd=dfd, search=search)


class FStat(ECcone):
    """ EC densities for a F random field.
    """
    def __init__(self, dfn, dfd=np.inf, search=[1]):
        self.dfn = dfn
        ECcone.__init__(self, mu=spherical_search(self.dfn), search=search, dfd=dfd)

    def __call__(self, x, search=None):
        return ECcone.__call__(self, np.sqrt(x * self.dfn), search=search)


class Roy(ECcone):
    """ Roy's maximum root

    Maximize an F_{dfd,dfn} statistic over a sphere of dimension k.
    """
    def __init__(self, dfn=1, dfd=np.inf, k=1, search=[1]):
        product = spherical_search(k)
        self.k = k
        self.dfn = dfn
        ECcone.__init__(self, mu=spherical_search(self.dfn),
                        search=search, dfd=dfd, product=product)

    def __call__(self, x, search=None):
        return ECcone.__call__(self, np.sqrt(x * self.dfn), search=search)


class MultilinearForm(ECcone):
    """ Maximize a multivariate Gaussian form

    Maximized over spheres of dimension dims. See:

    Kuri, S. & Takemura, A. (2001).
    'Tail probabilities of the maxima of multilinear forms and
    their applications.' Ann. Statist. 29(2): 328-371.
    """
    def __init__(self, *dims, **keywords):
        product = IntrinsicVolumes([1])

        if keywords.has_key('search'):
            search = keywords['search']
        else:
            search = [1]

        for d in dims:
            product *= spherical_search(d)
        product.mu /= 2.**(len(dims)-1)

        ECcone.__init__(self, search=search, product=product)


class Hotelling(ECcone):
    """ Hotelling's T^2

    Maximize an F_{1,dfd}=T_dfd^2 statistic over a sphere of dimension
    `k`.
    """
    def __init__(self, dfd=np.inf, k=1, search=[1]):
        product = spherical_search(k)
        self.k = k
        ECcone.__init__(self, mu=[1], search=search, dfd=dfd, product=product)

    def __call__(self, x, search=None):
        return ECcone.__call__(self, np.sqrt(x), search=search)


class OneSidedF(ECcone):
    """ EC densities for one-sided F statistic

    See:

    Worsley, K.J. & Taylor, J.E. (2005). 'Detecting fMRI activation
    allowing for unknown latency of the hemodynamic response.'
    Neuroimage, 29,649-654.
    """
    def __init__(self, dfn, dfd=np.inf, search=[1]):
        self.dfn = dfn
        self.regions = [spherical_search(dfn), spherical_search(dfn-1)]
        ECcone.__init__(self, mu=spherical_search(self.dfn), search=search, dfd=dfd)

    def __call__(self, x, search=None):
        IntrinsicVolumes.__init__(self, self.regions[0])
        d1 = ECcone.__call__(self, np.sqrt(x * self.dfn), search=search)
        IntrinsicVolumes.__init__(self, self.regions[1])
        d2 = ECcone.__call__(self, np.sqrt(x * (self.dfn-1)), search=search)
        self.mu = self.regions[0].mu
        return (d1 - d2) * 0.5


class ChiBarSquared(ChiSquared):
    def _getmu(self):
        x = np.linspace(0, 2 * self.dfn, 100)
        sf = 0.
        g = Gaussian()
        for i in range(1, self.dfn+1):
            sf += binomial(self.dfn, i) * stats.chi.sf(x, i) / np.power(2., self.dfn) 

        d = np.array([g.density(np.sqrt(x), j) for j in range(self.dfn)])
        c = np.dot(pinv(d.T), sf)
        sf += 1. / np.power(2, self.dfn)
        self.mu = IntrinsicVolumes(c)

    def __init__(self, dfn=1, search=[1]):
        ChiSquared.__init__(self, dfn=dfn, search=search)
        self._getmu()

    def __call__(self, x, dim=0, search=[1]):
        if search is None:
            search = self.stat
        else:
            search = IntrinsicVolumes(search) * self.stat
        return FStat.__call__(self, x, dim=dim, search=search)


def scale_space(region, interval, kappa=1.):
    """ scale space intrinsic volumes of region x interval

    See:

    Siegmund, D.O and Worsley, K.J. (1995). 'Testing for a signal
    with unknown location and scale in a stationary Gaussian random
    field.'  Annals of Statistics, 23:608-639.

    and

    Taylor, J.E. & Worsley, K.J. (2005). 'Random fields of multivariate
    test statistics, with applications to shape analysis and fMRI.'

    (available on http://www.math.mcgill.ca/keith
    """
    w1, w2 = interval
    region = IntrinsicVolumes(region)

    D = region.order
 
    out = np.zeros((D+2,))

    out[0] = region.mu[0]
    for i in range(1, D+2):
        if i < D+1:
            out[i] = (1./w1 + 1./w2) * region.mu[i] * 0.5
        for j in range(int(np.floor((D-i+1)/2.)+1)):
            denom = (i + 2*j - 1.)
            # w^-i/i when i=0 
            # according to Keith Worsley the 2005 paper has a typo
            if denom == 0:
                f = np.log(w2/w1) 
            else:
                f = (w1**(-i-2*j+1) - w2**(-i-2*j+1)) / denom
            f *= kappa**((1-2*j)/2.) * (-1)**j * factorial(int(denom))
            f /= (1 - 2*j) * (4*np.pi)**j * factorial(j) * factorial(i-1)
            out[i] += region.mu[int(denom)] * f
    return IntrinsicVolumes(out)



#!/usr/bin/env python
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
This modules launches a one-sample test on a dataset
Statistical significance is obtained using cluster-level inference
and permutation testing.

Author: Alexis Roche, Bertrand Thirion 2009-2012
"""
import numpy as np
from nibabel import Nifti1Image as Image

import nipy.labs.statistical_mapping as sm
from nipy.utils import example_data


def remake_images():
    # Get group data
    group_data = example_data.get_filename(
        'neurospin', 'language_babies', 'offset_002.npz')
    f = np.load(group_data)
    data, vardata, xyz = f['mat'], f['var'], f['xyz']
    dX = xyz[0].max() + 1
    dY = xyz[1].max() + 1
    dZ = xyz[2].max() + 1
    aux = np.zeros([dX, dY, dZ])
    data_images = []
    vardata_images = []
    mask_images = []
    for i in range(data.shape[0]):
        aux[list(xyz)] = data[i]
        data_images.append(Image(aux.copy(), np.eye(4)))
        aux[list(xyz)] = vardata[i]
        vardata_images.append(Image(aux.copy(), np.eye(4)))
        aux[list(xyz)] = 1
        mask_images.append(aux)

    return data_images, vardata_images, mask_images

data_images, vardata_images, mask_images = remake_images()

zimg, mask, nulls = sm.onesample_test(data_images,
                                      None,
                                      mask_images,
                                      'wilcoxon',
                                      permutations=1024,
                                      cluster_forming_th=0.01)
clusters, info = sm.cluster_stats(zimg, mask, 0.01, nulls=nulls)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from .linalg import (blas_dnrm2, blas_dasum, blas_ddot, blas_daxpy, blas_dscal,
                     blas_dgemm, blas_dsymm, blas_dtrmm, blas_dtrsm, blas_dsyrk,
                     blas_dsyr2k, matrix_add, matrix_get, matrix_transpose,
                     vector_get, vector_set, vector_add, vector_sub, vector_mul,
                     vector_div, vector_sum)
from .array import (array_get, array_get_block, array_add, array_sub, array_mul,
                    array_div)
from .wrapper import (c_types, fff_type, npy_type, copy_vector, pass_matrix,
                      pass_vector, pass_array, pass_vector_via_iterator,
                      sum_via_iterators, copy_via_iterators)

from nipy.testing import Tester
test = Tester().test
bench = Tester().bench



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
3D visualization of activation maps using Mayavi

"""

# Author: Gael Varoquaux <gael dot varoquaux at normalesup dot org>
# License: BSD

import os
import tempfile

# Standard scientific libraries imports (more specific imports are
# delayed, so that the part module can be used without them).
import numpy as np
from scipy import stats

# Local imports
from .anat_cache import mni_sform, mni_sform_inv, _AnatCache
from .coord_tools import coord_transform

# A module global to avoid creating multiple time an offscreen engine.
off_screen_engine = None

################################################################################
# Helper functions
def affine_img_src(data, affine, scale=1, name='AffineImage',
                reverse_x=False):
    """ Make a Mayavi source defined by a 3D array and an affine, for
        wich the voxel of the 3D array are mapped by the affine.
        
        Parameters
        -----------
        data: 3D ndarray
            The data arrays
        affine: (4 x 4) ndarray
            The (4 x 4) affine matrix relating voxels to world
            coordinates.
        scale: float, optional
            An optional addition scaling factor.
        name: string, optional
            The name of the Mayavi source created.
        reverse_x: boolean, optional
            Reverse the x (lateral) axis. Useful to compared with
            images in radiologic convention.

        Notes
        ------
        The affine should be diagonal.
    """
    # Late import to avoid triggering wx imports before needed.
    try:
        from mayavi.sources.api import ArraySource
    except ImportError:
        # Try out old install of Mayavi, with namespace packages
        from enthought.mayavi.sources.api import ArraySource
    center = np.r_[0, 0, 0, 1]
    spacing = np.diag(affine)[:3]
    origin = np.dot(affine, center)[:3]
    if reverse_x:
        # Radiologic convention
        spacing[0] *= -1
        origin[0] *= -1
    src = ArraySource(scalar_data=np.asarray(data, dtype=np.float), 
                           name=name,
                           spacing=scale*spacing,
                           origin=scale*origin)
    return src 


################################################################################
# Mayavi helpers
def autocrop_img(img, bg_color):
    red, green, blue = bg_color

    outline =  ( (img[..., 0] != red)
                +(img[..., 1] != green)
                +(img[..., 2] != blue)
                )
    outline_x = outline.sum(axis=0)
    outline_y = outline.sum(axis=1)

    outline_x = np.where(outline_x)[0]
    outline_y = np.where(outline_y)[0]

    if len(outline_x) == 0:
        return img
    else:
        x_min = outline_x.min()
        x_max = outline_x.max()
    if len(outline_y) == 0:
        return img
    else:
        y_min = outline_y.min()
        y_max = outline_y.max()
    return img[y_min:y_max, x_min:x_max]



def m2screenshot(mayavi_fig=None, mpl_axes=None, autocrop=True):
    """ Capture a screeshot of the Mayavi figure and display it in the
        matplotlib axes.
    """
    import pylab as pl
    # Late import to avoid triggering wx imports before needed.
    try:
        from mayavi import mlab
    except ImportError:
        # Try out old install of Mayavi, with namespace packages
        from enthought.mayavi import mlab

    if mayavi_fig is None:
        mayavi_fig = mlab.gcf()
    else:
        mlab.figure(mayavi_fig)
    if mpl_axes is not None:
        pl.axes(mpl_axes)

    filename = tempfile.mktemp('.png')
    mlab.savefig(filename, figure=mayavi_fig)
    image3d = pl.imread(filename)
    if autocrop:
        bg_color = mayavi_fig.scene.background
        image3d = autocrop_img(image3d, bg_color)
    pl.imshow(image3d)
    pl.axis('off')
    os.unlink(filename)
    # XXX: Should switch back to previous MPL axes: we have a side effect
    # here.


################################################################################
# Anatomy outline
################################################################################

def plot_anat_3d(anat=None, anat_affine=None, scale=1,
                 sulci_opacity=0.5, gyri_opacity=0.3,
                 opacity=None,
                 skull_percentile=78, wm_percentile=79,
                 outline_color=None):
    """ 3D anatomical display

    Parameters
    ----------
    skull_percentile : float, optional
        The percentile of the values in the image that delimit the skull from
        the outside of the brain. The smaller the fraction of you field of view
        is occupied by the brain, the larger this value should be.
    wm_percentile : float, optional
        The percentile of the values in the image that delimit the white matter
        from the grey matter. Typical this is skull_percentile + 1
    """
    # Late import to avoid triggering wx imports before needed.
    try:
        from mayavi import mlab
    except ImportError:
        # Try out old install of Mayavi, with namespace packages
        from enthought.mayavi import mlab
    fig = mlab.gcf()
    disable_render = fig.scene.disable_render
    fig.scene.disable_render = True
    if anat is None:
        anat, anat_affine, anat_max = _AnatCache.get_anat()
        anat_blurred = _AnatCache.get_blurred()
        skull_threshold = 4800
        inner_threshold = 5000
        upper_threshold = 7227.8
    else:
        from scipy import ndimage
        # XXX: This should be in a separate function
        voxel_size = np.sqrt((anat_affine[:3, :3]**2).sum()/3.)
        skull_threshold = stats.scoreatpercentile(anat.ravel(), 
                skull_percentile)
        inner_threshold = stats.scoreatpercentile(anat.ravel(), 
                wm_percentile)
        upper_threshold = anat.max()
        anat_blurred = ndimage.gaussian_filter(
                            (ndimage.morphology.binary_fill_holes(
                                ndimage.gaussian_filter(
                                    (anat > skull_threshold).astype(np.float), 
                                    6./voxel_size)
                                    > 0.5
                                )).astype(np.float),
                            2./voxel_size).T.ravel()

    if opacity is None:
        try:
            from tvtk.api import tvtk
        except ImportError:
            # Try out old install of Mayavi, with namespace packages
            from enthought.tvtk.api import tvtk
        version = tvtk.Version()
        if (version.vtk_major_version, version.vtk_minor_version) < (5, 2):
            opacity = .99
        else:
            opacity = 1
    ###########################################################################
    # Display the cortical surface (flattenned)
    anat_src = affine_img_src(anat, anat_affine, scale=scale, name='Anat')
    
    anat_src.image_data.point_data.add_array(anat_blurred)
    anat_src.image_data.point_data.get_array(1).name = 'blurred'
    anat_src.image_data.point_data.update()
    anat_blurred = mlab.pipeline.set_active_attribute(
                                anat_src, point_scalars='blurred')

    anat_blurred.update_pipeline()
    # anat_blurred = anat_src
    
    cortex_surf = mlab.pipeline.set_active_attribute(
                            mlab.pipeline.contour(anat_blurred), 
                    point_scalars='scalar')
        
    # XXX: the choice in vmin and vmax should be tuned to show the
    # sulci better
    cortex = mlab.pipeline.surface(cortex_surf,
                colormap='copper', 
                opacity=opacity,
                vmin=skull_threshold, 
                vmax=inner_threshold)
    cortex.enable_contours = True
    cortex.contour.filled_contours = True
    cortex.contour.auto_contours = False
    cortex.contour.contours = [0, inner_threshold, upper_threshold]
    #cortex.actor.property.backface_culling = True
    # XXX: Why do we do 'frontface_culling' to see the front.
    cortex.actor.property.frontface_culling = True

    cortex.actor.mapper.interpolate_scalars_before_mapping = True
    cortex.actor.property.interpolation = 'flat'

    # Add opacity variation to the colormap
    cmap = cortex.module_manager.scalar_lut_manager.lut.table.to_array()
    cmap[128:, -1] = gyri_opacity*255
    cmap[:128, -1] = sulci_opacity*255
    cortex.module_manager.scalar_lut_manager.lut.table = cmap

    if outline_color is not None:
        outline = mlab.pipeline.iso_surface(
                            anat_blurred,
                            contours=[0.4],
                            color=outline_color, 
                            opacity=.9)
        outline.actor.property.backface_culling = True


    fig.scene.disable_render = disable_render
    return cortex
 

################################################################################
# Maps
################################################################################

def plot_map_3d(map, affine, cut_coords=None, anat=None, anat_affine=None,
    threshold=None, offscreen=False, vmin=None, vmax=None, cmap=None,
    view=(38.5, 70.5, 300, (-2.7, -12, 9.1)),
    ):
    """ Plot a 3D volume rendering view of the activation, with an
        outline of the brain.

        Parameters
        ----------
        map : 3D ndarray
            The activation map, as a 3D image.
        affine : 4x4 ndarray
            The affine matrix going from image voxel space to MNI space.
        cut_coords: 3-tuple of floats, optional
            The MNI coordinates of a 3D cursor to indicate a feature
            or a cut, in MNI coordinates and order.
        anat : 3D ndarray, optional
            The anatomical image to be used as a background. If None, the
            MNI152 T1 1mm template is used. If False, no anatomical
            image is used.
        anat_affine : 4x4 ndarray, optional
            The affine matrix going from the anatomical image voxel space to 
            MNI space. This parameter is not used when the default 
            anatomical is used, but it is compulsory when using an
            explicite anatomical image.
        threshold : float, optional
            The lower threshold of the positive activation. This
            parameter is used to threshold the activation map.
        offscreen: boolean, optional
            If True, Mayavi attempts to plot offscreen. Will work only
            with VTK >= 5.2.
        vmin : float, optional
            The minimal value, for the colormap
        vmax : float, optional
            The maximum value, for the colormap
        cmap : a callable, or a pylab colormap
            A callable returning a (n, 4) array for n values between
            0 and 1 for the colors. This can be for instance a pylab
            colormap.

        Notes
        -----

        If you are using a VTK version below 5.2, there is no way to
        avoid opening a window during the rendering under Linux. This is
        necessary to use the graphics card for the rendering. You must
        maintain this window on top of others and on the screen.

    """
    # Late import to avoid triggering wx imports before needed.
    try:
        from mayavi import mlab
    except ImportError:
        # Try out old install of Mayavi, with namespace packages
        from enthought.mayavi import mlab
    if offscreen:
        global off_screen_engine
        if off_screen_engine is None:
            try:
                from mayavi.core.off_screen_engine import OffScreenEngine
            except ImportError:
                # Try out old install of Mayavi, with namespace packages
                from enthought.mayavi.core.off_screen_engine import OffScreenEngine
            off_screen_engine = OffScreenEngine()
        off_screen_engine.start()
        fig = mlab.figure('__private_plot_map_3d__', 
                                bgcolor=(1, 1, 1), fgcolor=(0, 0, 0),
                                size=(400, 330),
                                engine=off_screen_engine)
        mlab.clf(figure=fig)
    else:
        fig = mlab.gcf()
        fig = mlab.figure(fig, bgcolor=(1, 1, 1), fgcolor=(0, 0, 0),
                                                     size=(400, 350))
    disable_render = fig.scene.disable_render
    fig.scene.disable_render = True
    if threshold is None:
        threshold = stats.scoreatpercentile(
                                np.abs(map).ravel(), 80)
    contours = []
    lower_map = map[map <= -threshold]
    if np.any(lower_map):
        contours.append(lower_map.max())
    upper_map = map[map >= threshold]
    if np.any(upper_map):
        contours.append(map[map > threshold].min())


    ###########################################################################
    # Display the map using iso-surfaces
    if len(contours) > 0:
        map_src = affine_img_src(map, affine)
        module = mlab.pipeline.iso_surface(map_src,
                                        contours=contours,
                                        vmin=vmin, vmax=vmax)
        if hasattr(cmap, '__call__'):
            # Stick the colormap in mayavi
            module.module_manager.scalar_lut_manager.lut.table \
                    = (255*cmap(np.linspace(0, 1, 256))).astype(np.int)
    else:
        module = None

    if not anat is False:
        plot_anat_3d(anat=anat, anat_affine=anat_affine, scale=1.05,
                     outline_color=(.9, .9, .9),
                     gyri_opacity=.2)

    ###########################################################################
    # Draw the cursor
    if cut_coords is not None:
        x0, y0, z0 = cut_coords
        mlab.plot3d((-90, 90), (y0, y0), (z0, z0), 
                    color=(.5, .5, .5), tube_radius=0.25)
        mlab.plot3d((x0, x0), (-126, 91), (z0, z0), 
                    color=(.5, .5, .5), tube_radius=0.25)
        mlab.plot3d((x0, x0), (y0, y0), (-72, 109), 
                            color=(.5, .5, .5), tube_radius=0.25)

    mlab.view(*view)
    fig.scene.disable_render = disable_render
    
    return module


def demo_plot_map_3d():
    map = np.zeros((182, 218, 182))
    # Color a asymetric rectangle around Broca area:
    x, y, z = -52, 10, 22
    x_map, y_map, z_map = coord_transform(x, y, z, mni_sform_inv)
    map[x_map-5:x_map+5, y_map-3:y_map+3, z_map-10:z_map+10] = 1
    plot_map_3d(map, mni_sform, cut_coords=(x, y, z))





""" Testing
"""

import numpy as np

from .._registration import (_cspline_transform,
                             _cspline_sample1d,
                             _cspline_sample4d)

from numpy.testing import assert_array_almost_equal

from nose.tools import assert_true, assert_equal, assert_raises


def test_sample1d():
    a = np.random.rand(100)
    c = _cspline_transform(a)
    x = np.arange(100)
    b = np.zeros(100)
    b = _cspline_sample1d(b, c, x)
    assert_array_almost_equal(a, b)
    b = _cspline_sample1d(b, c, x, mode='nearest')
    assert_array_almost_equal(a, b)


def test_sample4d():
    a = np.random.rand(4, 5, 6, 7)
    c = _cspline_transform(a)
    x = np.mgrid[0:4, 0:5, 0:6, 0:7]
    b = np.zeros(a.shape)
    args = list(x)
    b = _cspline_sample4d(b, c, *args)
    assert_array_almost_equal(a, b)
    args = list(x) + ['nearest' for i in range(4)]
    b = _cspline_sample4d(b, c, *args)
    assert_array_almost_equal(a, b)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Test the fast distance estimator
"""
import numpy as np
from numpy.testing import assert_almost_equal

from ..fast_distance import euclidean_distance as ed 

def test_euclidean_1():
    """ test that the euclidean distance is as expected
    """  
    nx, ny = (10, 12)
    X = np.random.randn(nx, 2)
    Y = np.random.randn(ny, 2)
    ED = ed(X, Y)
    ref = np.zeros((nx, ny))
    for i in range(nx):
    	ref[i] = np.sqrt(np.sum((Y - X[i])**2, 1))
	
    assert_almost_equal(ED, ref)		

 
def test_euclidean_2():
    """ test that the euclidean distance is as expected
    """  
    nx = 10
    X = np.random.randn(nx, 2)
    ED = ed(X)
    ref = np.zeros((nx, nx))
    for i in range(nx):
    	ref[i] = np.sqrt(np.sum((X - X[i])**2, 1))
	
    assert_almost_equal(ED, ref) 

  
if __name__ == "__main__":
    import nose
    nose.run(argv=['', __file__])
